# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional


class HeapElement(object):
    """
    An element of one of the heaps on SpiNNaker.
    """

    __slots__ = [
        # A pointer to the block
        "_block_address",
        # A pointer to the next block
        "_next_address",
        # True if the block is free
        "_is_free",
        # The tag of the block
        "_tag",
        # The app ID of the block
        "_app_id"
    ]

    def __init__(self, block_address: int, next_address: int, free: int):
        """
        :param int block_address: The address of this element on the heap
        :param int next_address: The address of the next element on the heap
        :param int free: The "free" element of the block as read from the heap
        """
        self._block_address = block_address
        self._next_address = next_address
        self._is_free = (free & 0xFFFF0000) != 0xFFFF0000
        self._tag: Optional[int] = None
        self._app_id: Optional[int] = None
        if not self._is_free:
            self._tag = free & 0xFF
            self._app_id = (free >> 8) & 0xFF

    @property
    def block_address(self) -> int:
        """
        The address of the block.

        :rtype: int
        """
        return self._block_address

    @property
    def next_address(self) -> int:
        """
        The address of the next block, or 0 if none.

        :rtype: int
        """
        return self._next_address

    @property
    def size(self) -> int:
        """
        The usable size of this block (not including the header).

        :rtype: int
        """
        return self._next_address - self._block_address - 8

    @property
    def is_free(self) -> bool:
        """
        Whether this block is a free block.

        :rtype: bool
        """
        return self._is_free

    @property
    def tag(self) -> Optional[int]:
        """
        The tag of the block if allocated, or `None` if not.

        :rtype: int or None
        """
        return self._tag

    @property
    def app_id(self) -> Optional[int]:
        """
        The application ID of the block if allocated, or `None` if not.

        :rtype: int or None
        """
        return self._app_id

    def __str__(self) -> str:
        if self._is_free:
            return f"FREE  0x{self._block_address:8X} SIZE: {self.size:9d}"
        assert self._tag is not None
        assert self._app_id is not None
        return "BLOCK 0x{self._block_address:8X} SIZE: {self.size:9d} " \
               "TAG: {self._tag:3d} APP_ID: {self._app_id:3d}"



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is a script used to check the state of a SpiNNaker machine.
"""

import argparse
import sys

from spinn_utilities.config_holder import set_config
from spinn_machine import CoreSubsets, CoreSubset

from spinnman.board_test_configuration import BoardTestConfiguration
from spinnman.config_setup import unittest_setup
from spinnman.model.enums import CPUState
from spinnman.transceiver import create_transceiver_from_hostname, Transceiver

SCAMP_ID = 0
IGNORED_IDS = {SCAMP_ID, 16}  # WHY 16?


def get_cores_in_run_state(txrx, app_id, print_all_chips):
    """
    :param Transceiver txrx:
    :param int app_id:
    :param bool print_all_chips:
    """
    count_finished = txrx.get_core_state_count(app_id, CPUState.FINISHED)
    count_run = txrx.get_core_state_count(app_id, CPUState.RUNNING)
    print(f'running: {count_run} finished: {count_finished}')

    machine = txrx.get_machine_details()
    print(f'machine width: {machine.width} height: {machine.height}')
    if print_all_chips:
        print(f'machine chips: {list(machine.chips)}')

    all_cores = []
    for chip in machine.chips:
        all_cores.append(CoreSubset(
            chip.x, chip.y, chip.placable_processors_ids))

    all_cores = CoreSubsets(core_subsets=all_cores)

    cpu_infos = txrx.get_cpu_infos(
        all_cores,
        [CPUState.FINISHED, CPUState.RUNNING, CPUState.WATCHDOG], True)
    cores_finished = cpu_infos.infos_for_state(CPUState.FINISHED)
    cores_running = cpu_infos.infos_for_state(CPUState.RUNNING)
    cores_watchdog = cpu_infos.infos_for_state(CPUState.WATCHDOG)

    for (x, y, p), _ in cores_running:
        if p not in IGNORED_IDS:
            print(f'run core: {x} {y} {p}')

    for (x, y, p), _ in cores_finished:
        print(f'finished core: {x} {y} {p}')

    for (x, y, p), _ in cores_watchdog:
        print(f'watchdog core: {x} {y} {p}')


def _make_transceiver(host, version, bmp_names) -> Transceiver:
    """
    :param host:
        Host to use or `None` to use test configuration for all parameters
    :type host: str or None
    :param version: Board version to use (`None` defaults to 5 unless host is
        192.168.240.253 (spin 3)
    :type version: int or None
    :param bmp_names: names of BMP connection
        or `None` to auto detect (if applicable)
    :type bmp_names: str or None
    :rtype: Transceiver
    """
    if host is None:
        config = BoardTestConfiguration()
        config.set_up_remote_board()
        host = config.remotehost
        bmp_names = None
        auto_detect_bmp = config.auto_detect_bmp
    else:
        if version is None:
            if host == "192.168.240.253":
                version = 3
            else:
                version = 5
        auto_detect_bmp = False
        set_config("Machine", "version", version)

    print(f"talking to SpiNNaker system at {host}")
    return create_transceiver_from_hostname(
        host, bmp_connection_data=bmp_names,
        auto_detect_bmp=auto_detect_bmp)


def main(args):
    """
    Runs the script.
    """
    unittest_setup()
    ap = argparse.ArgumentParser(
        description="Check the state of a SpiNNaker machine.")
    ap.add_argument(
        "-a", "--appid", help="the application ID to check", type=int,
        default=17)
    ap.add_argument(
        "-v", "--version", help="the version of your boards", type=int,
        default=None)
    ap.add_argument(
        "-b", "--bmp_names",
        help="the hostname or IP address of the BMP of the SpiNNaker machine "
             "to inspects",
        type=str, default=None)
    ap.add_argument(
        "-n", "--noprintchips", action="store_true", default=False,
        help=("don't print all the chips out; avoids a great deal of "
              "output for large machines"))
    ap.add_argument(
        "host", default=None, nargs='?',
        help="the hostname or IP address of the SpiNNaker machine to inspect")
    args = ap.parse_args(args)
    # These ought to be parsed from command line arguments
    app_id = args.appid
    print_chips = not args.noprintchips

    transceiver = _make_transceiver(args.host, args.version, args.bmp_names)
    try:
        get_cores_in_run_state(transceiver, app_id, print_chips)
    finally:
        transceiver.close()


if __name__ == "__main__":  # pragma: no cover
    main(sys.argv[1:])



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Sequence, Tuple
from typing_extensions import TypeAlias
from spinnman.model.enums import CPUState, RunTimeError, MailboxCommand

#: Size of `vcpu_t` in SARK.
CPU_INFO_BYTES = 128
CPU_USER_0_START_ADDRESS = 112
#: Offset into data of byte of processor state field.
STATE_FIELD_OFFSET = 48

#: Corresponds to vcpu_t in sark.h
_VCPU_PATTERN = struct.Struct("< 32s 3I 2B 2B 2I 2B H 3I 16s 2I 16x 4I")
_vcpu_t: TypeAlias = Tuple[
    # pylint: disable=wrong-spelling-in-comment
    bytes,             # 32s - r0-r7
    int, int, int,     # 3I  - psr, sp, lr
    int, int,          # 2B  - RT error code, Physical CPU
    int, int,          # 2B  - CPU state, Application ID
    int, int,          # 2I  - mbox msg MP->AP, mbox msg AP->MP
    int, int,          # 2B  - mbox command MP->AP, mbox command AP->MP
    int,               # H   - SW error count (saturating)
    int, int, int,     # 3I  - SW source filename, Source line, Time of loading
    bytes,             # 16s - Application name
    int, int,          # 2I  - IO buffer in SDRAM (or 0)
    #                    16x - Padding
    int, int, int, int]  # 4I- User0, User1, User2, User3
_REGISTERS_PATTERN = struct.Struct("<IIIIIIII")


class CPUInfo(object):
    """
    Represents information about the state of a CPU.

    This is the content of the `vcpu_t` for the processor, maintained by SARK.
    """
    __slots__ = [
        "__application_id",
        "__application_mailbox_command",
        "__app_mailbox",
        "__application_name",
        "__iobuf_address",
        "__link_register",
        "__monitor_mailbox_command",
        "__monitor_mailbox",
        "__physical_cpu_id",
        "__processor_state_register",
        "__registers",
        "__run_time_error",
        "__run_time_error_value",
        "__software_error_count",
        "__filename_address",
        "__line_number",
        "__software_version",
        "__stack_pointer",
        "__state",
        "__time",
        "__user",
        "__x", "__y", "__p"]

    def __init__(self, x: int, y: int, p: int, cpu_data: _vcpu_t):
        """
        :param int x: The x-coordinate of a chip
        :param int y: The y-coordinate of a chip
        :param int p: The ID of a core on the chip
        :param tuple cpu_data: A byte-string received from SDRAM on the board
        """
        # pylint: disable=too-many-arguments
        self.__x, self.__y, self.__p = x, y, p

        (registers,  # 32s 0
         self.__processor_state_register, self.__stack_pointer,
         self.__link_register,  run_time_error,
         self.__physical_cpu_id,  # 2B  44
         state, self.__application_id,  # 2B  46
         self.__app_mailbox, self.__monitor_mailbox,  # 2I  48
         app_mailbox_cmd, mon_mailbox_cmd,  # 2B  56
         self.__software_error_count,  # H   58
         self.__filename_address, self.__line_number, self.__time,  # 3I  60
         app_name,  # 16s 72
         self.__iobuf_address, self.__software_version,  # 2I  88
         # skipped                                                  # 16x 96
         user0, user1, user2, user3                                 # 4I  112
         ) = cpu_data

        index = app_name.find(b'\0')
        if index != -1:
            app_name = app_name[:index]
        self.__application_name = app_name.decode('ascii')

        self.__registers: Sequence[int] = _REGISTERS_PATTERN.unpack(registers)
        self.__run_time_error_value = run_time_error
        try:
            self.__run_time_error = RunTimeError(run_time_error)
        except ValueError:
            self.__run_time_error = RunTimeError.UNRECOGNISED
        self.__state = CPUState(state)

        self.__application_mailbox_command = MailboxCommand(app_mailbox_cmd)
        self.__monitor_mailbox_command = MailboxCommand(mon_mailbox_cmd)
        self.__user = (user0, user1, user2, user3)

    @property
    def x(self) -> int:
        """
        The X-coordinate of the chip containing the core.

        :return: The x-coordinate of the chip
        :rtype: int
        """
        return self.__x

    @property
    def y(self) -> int:
        """
        The y-coordinate of the chip containing the core.

        :return: The y-coordinate of the chip
        :rtype: int
        """
        return self.__y

    @property
    def p(self) -> int:
        """
        The ID of the core on the chip.

        :return: The ID of the core
        :rtype: int
        """
        return self.__p

    @property
    def state(self) -> CPUState:
        """
        The current state of the core.

        :return: The state of the core
        :rtype: CPUState
        """
        return self.__state

    @property
    def physical_cpu_id(self) -> int:
        """
        The physical ID of this processor.

        :return: The physical ID of the processor
        :rtype: int
        """
        return self.__physical_cpu_id

    @property
    def application_name(self) -> str:
        """
        The name of the application running on the core.

        :return: The name of the application
        :rtype: str
        """
        return self.__application_name

    @property
    def application_id(self) -> int:
        """
        The ID of the application running on the core.

        :return: The ID of the application
        :rtype: int
        """
        return self.__application_id

    @property
    def time(self) -> int:
        """
        The time at which the application started.

        :return: The time in seconds since 00:00:00 on the 1st January 1970
        :rtype: int
        """
        return self.__time

    @property
    def run_time_error(self) -> RunTimeError:
        """
        The reason for a run time error.

        :return: The run time error
        :rtype: RunTimeError
        """
        return self.__run_time_error

    @property
    def application_mailbox_command(self) -> MailboxCommand:
        """
        The command currently in the mailbox being sent from the monitor
        processor to the application.

        :return: The command
        :rtype: MailboxCommand
        """
        return self.__application_mailbox_command

    @property
    def application_mailbox_data_address(self) -> int:
        """
        The address of the data in SDRAM for the application mailbox.

        :return: The address of the data
        :rtype: int
        """
        return self.__app_mailbox

    @property
    def monitor_mailbox_command(self) -> MailboxCommand:
        """
        The command currently in the mailbox being sent from the
        application to the monitor processor.

        :return: The command
        :rtype: MailboxCommand
        """
        return self.__monitor_mailbox_command

    @property
    def monitor_mailbox_data_address(self) -> int:
        """
        The address of the data in SDRAM of the monitor mailbox.

        :return: The address of the data
        :rtype: int
        """
        return self.__monitor_mailbox

    @property
    def software_error_count(self) -> int:
        """
        The number of software errors counted. Saturating.

        :return: The number of software errors
        :rtype: int
        """
        return self.__software_error_count

    @property
    def software_source_filename_address(self) -> int:
        """
        The address of the filename of the software source.

        :return: The filename address
        :rtype: int
        """
        return self.__filename_address

    @property
    def software_source_line_number(self) -> int:
        """
        The line number of the software source.

        :return: The line number
        :rtype: int
        """
        return self.__line_number

    @property
    def processor_state_register(self) -> int:
        """
        The value in the processor state register.

        :return: The processor state register value
        :rtype: int
        """
        return self.__processor_state_register

    @property
    def stack_pointer(self) -> int:
        """
        The current stack pointer value.

        :return: The stack pointer value
        :rtype: int
        """
        return self.__stack_pointer

    @property
    def link_register(self) -> int:
        """
        The current link register value.

        :return: The link register value
        :rtype: int
        """
        return self.__link_register

    @property
    def registers(self) -> Sequence[int]:
        """
        The current register values (r0 - r7).

        :return: An array of 8 values, one for each register
        :rtype: list(int)
        """
        return self.__registers

    @property
    def user(self) -> Sequence[int]:
        """
        The current user values (user0 - user3).

        :return: An array of 4 values, one for each user value
        :rtype: list(int)
        """
        return self.__user

    @property
    def iobuf_address(self) -> int:
        """
        The address of the IOBUF buffer in SDRAM.

        :return: The address
        :rtype: int
        """
        return self.__iobuf_address

    @property
    def software_version(self) -> int:
        """
        The software version.

        :return: The software version
        :rtype: int
        """
        return self.__software_version

    def __str__(self) -> str:
        return (f"{self.x}:{self.y}:{self.p:02n} ({self.physical_cpu_id:02n}) "
                f"{self.__state.name:18} {self.__application_name:16s} "
                f"{self.__application_id:3n}")

    def get_status_string(self) -> str:
        """
        Get a string indicating the status of the given core.

        :rtype: str
        """
        if self.state == CPUState.RUN_TIME_EXCEPTION:
            rte_string = f"{self.run_time_error.name}"
            if self.run_time_error == RunTimeError.UNRECOGNISED:
                rte_string += f" ({self.__run_time_error_value})"
            return (
                f"{self.__x}:{self.__y}:{self.__p} "
                f"(ph: {self.__physical_cpu_id}) "
                f"in state {self.__state.name}:{rte_string}\n"
                f"    r0={self.__registers[0]}, r1={self.__registers[1]}, "
                f"r2={self.__registers[2]}, r3={self.__registers[3]}\n"
                f"    r4={self.__registers[4]}, r5={self.__registers[5]}, "
                f"r6={self.__registers[6]}, r7={self.__registers[7]}\n"
                f"    PSR={self.__processor_state_register}, "
                f"SP={self.__stack_pointer}, LR={self.__link_register}\n")
        else:
            return (
                f"{self.__x}:{self.__y}:{self.__p} "
                f"in state {self.__state.name}\n")

    @staticmethod
    def mock_info(
            x: int, y: int, p: int, physical_cpu_id: int, state: CPUState):
        """
        Makes a CPU_info object for Testing purposes

        :param int x:
        :param int y:
        :param int p:
        :param int physical_cpu_id:
        :param CPUState CPIstate:
        """
        registers = b'@\x00\x07\x08\xff\x00\x00\x00\x00\x00\x80\x00\xad\x00' \
                    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                    b'\x00\x00\x00\x00\x00'
        time = 1687857627
        application_name = b'scamp-3\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        iobuff_address = 197634
        cpu_data = (
            registers, 0, 0, 0, 0, physical_cpu_id, state.value, 0, 0, 0, 0,
            0, 0, 0, 0, time, application_name, iobuff_address, 0, 0, 0, 0, 0)
        return CPUInfo(x, y, p, cpu_data)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Implementation of the client for the Spalloc web service.
"""
import os
import time
from logging import getLogger

import queue
import struct
import threading
from time import sleep
from typing import (Any, Callable, Dict, FrozenSet, Iterable, List, Mapping,
                    Optional, Tuple, cast)
from urllib.parse import urlparse, urlunparse, ParseResult

from packaging.version import Version
import requests
from typing_extensions import TypeAlias
from websocket import WebSocket  # type: ignore

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.abstract_context_manager import AbstractContextManager
from spinn_utilities.log import FormatAdapter
from spinn_utilities.typing.coords import XY
from spinn_utilities.typing.json import JsonObject, JsonValue
from spinn_utilities.overrides import overrides

from spinnman.connections.udp_packet_connections import UDPConnection
from spinnman.connections.abstract_classes import Connection, Listenable
from spinnman.constants import SCP_SCAMP_PORT, UDP_BOOT_CONNECTION_DEFAULT_PORT
from spinnman.exceptions import SpinnmanTimeoutException
from spinnman.exceptions import SpallocException
from spinnman.transceiver import (
    Transceiver, create_transceiver_from_connections)

from .abstract_spalloc_client import AbstractSpallocClient
from .proxy_protocol import ProxyProtocol
from .session import Session, SessionAware
from .spalloc_boot_connection import SpallocBootConnection
from .spalloc_eieio_connection import SpallocEIEIOConnection
from .spalloc_eieio_listener import SpallocEIEIOListener
from .spalloc_job import SpallocJob
from .spalloc_machine import SpallocMachine
from .spalloc_proxied_connection import SpallocProxiedConnection
from .spalloc_scp_connection import SpallocSCPConnection
from .spalloc_state import SpallocState
from .utils import parse_service_url, get_hostname

logger = FormatAdapter(getLogger(__name__))
_open_req = struct.Struct("<IIIII")
_close_req = struct.Struct("<III")
_open_listen_req = struct.Struct("<II")
# Open and close share the response structure
_open_close_res = struct.Struct("<III")
_open_listen_res = struct.Struct("<IIIBBBBI")
_msg = struct.Struct("<II")
_msg_to = struct.Struct("<IIIII")

KEEP_ALIVE_PERIOND = 120


def fix_url(url: Any) -> str:
    """
    Makes sure the url is the correct format.

    :param str url: original url
    :rtype: str
    """
    parts = urlparse(url)
    if parts.scheme != 'https':
        parts = ParseResult("https", parts.netloc, parts.path,
                            parts.params, parts. query, parts.fragment)
    if not parts.path.endswith("/"):
        parts = ParseResult(parts.scheme, parts.netloc, parts.path + "/",
                            parts.params, parts.query, parts.fragment)
    return urlunparse(parts)


class SpallocClient(AbstractContextManager, AbstractSpallocClient):
    """
    Basic client library for talking to new Spalloc.
    """
    __slots__ = ("__session",
                 "__machines_url", "__jobs_url", "version",
                 "__group", "__collab", "__nmpi_job", "__nmpi_user")

    def __init__(
            self, service_url: str,
            username: Optional[str] = None, password: Optional[str] = None,
            bearer_token: Optional[str] = None,
            group: Optional[str] = None, collab: Optional[str] = None,
            nmpi_job: Optional[int] = None, nmpi_user: Optional[str] = None):
        """
        :param str service_url: The reference to the service.
            May have username and password supplied as part of the network
            location; if so, the ``username`` and ``password`` arguments
            *must* be ``None``. If ``username`` and ``password`` are not given,
            not even within the URL, the ``bearer_token`` must be not ``None``.
        :param str username:
            The user name to use. If not provided nor in service_url
            environment variable SPALLOC_USER will be used.
        :param str password:
            The password to use. If not provided nor in service_url
            environment variable SPALLOC_PASSWORD will be used.
        :param str bearer_token: The bearer token to use
        """
        if username is None and password is None:
            service_url, username, password = parse_service_url(service_url)
        if username is None:
            username = os.getenv("SPALLOC_USER", None)
        if password is None:
            password = os.getenv("SPALLOC_PASSWORD", None)

        self.__session: Optional[Session] = Session(
            service_url, username, password, bearer_token)
        obj = self.__session.renew()
        v = cast(JsonObject, obj["version"])
        self.version = Version(
            f"{v['major-version']}.{v['minor-version']}.{v['revision']}")
        self.__machines_url = fix_url(obj["machines-ref"])
        self.__jobs_url = fix_url(obj["jobs-ref"])
        self.__group = group
        self.__collab = collab
        self.__nmpi_job = nmpi_job
        self.__nmpi_user = nmpi_user
        logger.info("established session to {} for {}", service_url, username)

    def get_job(self, job_id: str) -> SpallocJob:
        """
        Get a job by its job id.

        :param str job_id: The job id.
        :rtype: SpallocJob
        """
        assert self.__session
        return _SpallocJob(
            self.__session, fix_url(f"{self.__jobs_url}/{job_id}"))

    @staticmethod
    def open_job_from_database(
            service_url, job_url, cookies, headers) -> SpallocJob:
        """
        Create a job from the description in the attached database. This is
        intended to allow for access to the job's allocated resources from
        visualisers and other third party code participating in the SpiNNaker
        Tools Notification Protocol.

        .. note::
            The job is not verified to exist and be running. The session
            credentials may have expired; if so, the job will be unable to
            regenerate them.

        :param str service_url:
        :param str job_url:
        :param dict(str, str) cookies:
        :param dict(str, str) headers:

        :return:
            The job handle, or ``None`` if the records in the database are
            absent or incomplete.
        :rtype: SpallocJob
        """
        session = Session(service_url, session_credentials=(cookies, headers))
        return _SpallocJob(session, job_url)

    @overrides(AbstractSpallocClient.list_machines)
    def list_machines(self) -> Dict[str, SpallocMachine]:
        assert self.__session
        obj = self.__session.get(self.__machines_url).json()
        return {m["name"]: _SpallocMachine(self.__session, m)
                for m in obj["machines"]}

    @overrides(AbstractSpallocClient.list_jobs)
    def list_jobs(self, deleted: bool = False) -> Iterable[SpallocJob]:
        assert self.__session
        obj = self.__session.get(
            self.__jobs_url,
            deleted=("true" if deleted else "false")).json()
        while obj["jobs"]:
            for u in obj["jobs"]:
                yield _SpallocJob(self.__session, fix_url(u))
            if "next" not in obj:
                break
            obj = self.__session.get(obj["next"]).json()

    def _create(self, create: Mapping[str, JsonValue],
                machine_name: Optional[str]) -> SpallocJob:
        assert self.__session
        operation = dict(create)
        if machine_name:
            operation["machine-name"] = machine_name
        else:
            operation["tags"] = ["default"]
        if self.__group is not None:
            operation["group"] = self.__group
        if self.__collab is not None:
            operation["nmpi-collab"] = self.__collab
        if self.__nmpi_job is not None:
            operation["nmpi-job-id"] = self.__nmpi_job
            if self.__nmpi_user is not None:
                operation["owner"] = self.__nmpi_user
        logger.info("Posting {} to {}", operation, self.__jobs_url)
        r = self.__session.post(self.__jobs_url, operation, timeout=30)
        url = r.headers["Location"]
        return _SpallocJob(self.__session, fix_url(url))

    @overrides(AbstractSpallocClient.create_job)
    def create_job(
            self, num_boards: int = 1,
            machine_name: Optional[str] = None,
            keepalive: int = KEEP_ALIVE_PERIOND) -> SpallocJob:
        return self._create({
            "num-boards": int(num_boards),
            "keepalive-interval": f"PT{int(keepalive)}S"
        }, machine_name)

    @overrides(AbstractSpallocClient.create_job_rect)
    def create_job_rect(
            self, width: int, height: int,
            machine_name: Optional[str] = None,
            keepalive: int = KEEP_ALIVE_PERIOND) -> SpallocJob:
        return self._create({
            "dimensions": {
                "width": int(width),
                "height": int(height)
            },
            "keepalive-interval": f"PT{int(keepalive)}S"
        }, machine_name)

    @overrides(AbstractSpallocClient.create_job_board)
    def create_job_board(
            self, triad: Optional[Tuple[int, int, int]] = None,
            physical: Optional[Tuple[int, int, int]] = None,
            ip_address: Optional[str] = None,
            machine_name: Optional[str] = None,
            keepalive: int = KEEP_ALIVE_PERIOND) -> SpallocJob:
        board: JsonObject
        if triad:
            x, y, z = triad
            board = {"x": int(x), "y": int(y), "z": int(z)}
        elif physical:
            c, f, b = physical
            board = {"cabinet": int(c), "frame": int(f), "board": int(b)}
        elif ip_address:
            board = {"address": str(ip_address)}
        else:
            raise KeyError("at least one of triad, physical and ip_address "
                           "must be given")
        return self._create({
            "board": board,
            "keepalive-interval": f"PT{int(keepalive)}S"
        }, machine_name)

    @overrides(AbstractSpallocClient.create_job_rect_at_board)
    def create_job_rect_at_board(
            self, width: int, height: int,
            triad: Optional[Tuple[int, int, int]] = None,
            physical: Optional[Tuple[int, int, int]] = None,
            ip_address: Optional[str] = None,
            machine_name: Optional[str] = None,
            keepalive: int = KEEP_ALIVE_PERIOND,
            max_dead_boards: int = 0) -> SpallocJob:
        board: JsonObject
        if triad:
            x, y, z = triad
            board = {"x": int(x), "y": int(y), "z": int(z)}
        elif physical:
            c, f, b = physical
            board = {"cabinet": int(c), "frame": int(f), "board": int(b)}
        elif ip_address:
            board = {"address": str(ip_address)}
        else:
            raise KeyError("at least one of triad, physical and ip_address "
                           "must be given")
        return self._create({
            "dimensions": {
                "width": int(width),
                "height": int(height)
            },
            "board": board,
            "keepalive-interval": f"PT{int(keepalive)}S",
            "max-dead-boards": int(max_dead_boards)
        }, machine_name)

    def close(self) -> None:
        # pylint: disable=protected-access
        if self.__session is not None:
            self.__session._purge()
        self.__session = None


class _ProxyServiceError(IOError):
    """
    An error passed to us from the server over the proxy channel.
    """


class _SpallocMachine(SessionAware, SpallocMachine):
    """
    Represents a Spalloc-controlled machine.

    Don't make this yourself. Use :py:class:`SpallocClient` instead.
    """
    __slots__ = ("__name", "__tags", "__width", "__height",
                 "__dead_boards", "__dead_links")

    def __init__(self, session: Session, machine_data: JsonObject):
        """
        :param _Session session:
        :param dict machine_data:
        """
        super().__init__(session, cast(str, machine_data["uri"]))
        self.__name = cast(str, machine_data["name"])
        self.__tags = frozenset(cast(List[str], machine_data["tags"]))
        self.__width = cast(int, machine_data["width"])
        self.__height = cast(int, machine_data["height"])
        self.__dead_boards = cast(list, machine_data["dead-boards"])
        self.__dead_links = cast(list, machine_data["dead-links"])

    @property
    @overrides(SpallocMachine.name)
    def name(self) -> str:
        return self.__name

    @property
    @overrides(SpallocMachine.tags)
    def tags(self) -> FrozenSet[str]:
        return self.__tags

    @property
    @overrides(SpallocMachine.width)
    def width(self) -> int:
        return self.__width

    @property
    @overrides(SpallocMachine.height)
    def height(self) -> int:
        return self.__height

    @property
    @overrides(SpallocMachine.dead_boards)
    def dead_boards(self) -> list:
        return self.__dead_boards

    @property
    @overrides(SpallocMachine.dead_links)
    def dead_links(self) -> list:
        return self.__dead_links

    @property
    @overrides(SpallocMachine.area)
    def area(self) -> Tuple[int, int]:
        return (self.width, self.height)

    def __repr__(self):
        return "SpallocMachine" + str((
            self.name, self.tags, self.width, self.height, self.dead_boards,
            self.dead_links))


class _ProxyPing(threading.Thread):
    """
    Sends ping messages to an open websocket
    """

    def __init__(self, ws, sleep_time=30):
        super().__init__(daemon=True)
        self.__ws = ws
        self.__sleep_time = sleep_time
        self.__closed = False
        self.start()

    def run(self):
        """
        The handler loop of this thread
        """
        while self.__ws.connected:
            try:
                self.__ws.ping()
            except Exception:  # pylint: disable=broad-except
                # If closed, ignore error and get out of here
                if self.__closed:
                    break

                # Make someone aware of the error
                logger.exception("Error in websocket before close")
            sleep(self.__sleep_time)

    def close(self):
        """
        Mark as closed to avoid error messages.
        """
        self.__closed = True


_WSCB: TypeAlias = Callable[[Optional[bytes]], None]


class _ProxyReceiver(threading.Thread):
    """
    Receives all messages off an open websocket and dispatches them to
    registered listeners.
    """

    def __init__(self, ws: WebSocket):
        super().__init__(daemon=True)
        self.__ws = ws
        self.__returns: Dict[int, _WSCB] = {}
        self.__handlers: Dict[int, _WSCB] = {}
        self.__correlation_id = 0
        self.__closed = False
        self.start()

    def run(self) -> None:
        """
        The handler loop of this thread.
        """
        while self.__ws.connected:
            try:
                result: Tuple[int, bytes] = self.__ws.recv_data()
                frame = result[1]
                if len(frame) < _msg.size:
                    # Message is out of protocol
                    continue
            except Exception:  # pylint: disable=broad-except
                # If closed, ignore error and get out of here
                if self.__closed:
                    break

                # Make someone aware of the error
                logger.exception("Error in websocket before close")

                # If we are disconnected before closing, make errors happen
                if not self.__ws.connected:
                    for rt in self.__returns.values():
                        rt(None)
                    for hd in self.__handlers.values():
                        hd(None)
                    break
            code, num = _msg.unpack_from(frame, 0)
            if code == ProxyProtocol.MSG:
                self.dispatch_message(num, frame)
            else:
                self.dispatch_return(num, frame)

    def expect_return(self, handler: _WSCB) -> int:
        """
        Register a one-shot listener for a call-like message's return.

        :return: The correlation ID
        """
        c = self.__correlation_id
        self.__correlation_id += 1
        self.__returns[c] = handler
        return c

    def listen(self, channel_id: int, handler: _WSCB):
        """
        Register a persistent listener for one-way messages.
        """
        self.__handlers[channel_id] = handler

    def dispatch_return(self, correlation_id: int, msg: bytes):
        """
        Dispatch a received call-return message.
        """
        handler = self.__returns.pop(correlation_id, None)
        if handler:
            handler(msg)

    def dispatch_message(self, channel_id: int, msg: bytes):
        """
        Dispatch a received one-way message.
        """
        handler = self.__handlers.get(channel_id)
        if handler:
            handler(msg)

    def unlisten(self, channel_id: int):
        """
        Deregister a listener for a channel
        """
        self.__handlers.pop(channel_id)

    def close(self) -> None:
        """
        Mark receiver closed to avoid errors
        """
        self.__closed = True


class _SpallocJob(SessionAware, SpallocJob):
    """
    Represents a job in Spalloc.

    Don't make this yourself. Use :py:class:`SpallocClient` instead.
    """
    __slots__ = ("__machine_url", "__chip_url",
                 "_keepalive_url", "__proxy_handle",
                 "__proxy_thread", "__proxy_ping")

    def __init__(self, session: Session, job_handle: str):
        """
        :param _Session session:
        :param str job_handle:
        """
        super().__init__(session, job_handle)
        logger.info("established job at {}", job_handle)
        self.__machine_url = self._url + "machine"
        self.__chip_url = self._url + "chip"
        self._keepalive_url: Optional[str] = self._url + "keepalive"
        self.__proxy_handle: Optional[WebSocket] = None
        self.__proxy_thread: Optional[_ProxyReceiver] = None
        self.__proxy_ping: Optional[_ProxyPing] = None
        keep_alive = threading.Thread(
            target=self.__start_keepalive, daemon=True)
        keep_alive.start()

    @overrides(SpallocJob.get_session_credentials_for_db)
    def get_session_credentials_for_db(self) -> Mapping[Tuple[str, str], str]:
        config = {}
        config["SPALLOC", "service uri"] = self._service_url
        config["SPALLOC", "job uri"] = self._url
        cookies, headers = self._session_credentials
        if "Authorization" in headers:
            # We never write the authorisation headers themselves;
            # we just extend the session
            del headers["Authorization"]
        for k, v in cookies.items():
            config["COOKIE", k] = v
        for k, v in headers.items():
            config["HEADER", k] = v
        return config

    @overrides(SpallocJob.get_state)
    def get_state(self, wait_for_change: bool = False) -> SpallocState:
        timeout: Optional[int] = 10
        if wait_for_change:
            timeout = None
        obj = self._get(
            self._url, wait=wait_for_change, timeout=timeout).json()
        return SpallocState[obj["state"]]

    @overrides(SpallocJob.get_root_host)
    def get_root_host(self) -> Optional[str]:
        r = self._get(self.__machine_url)
        if r.status_code == 204:
            return None
        obj = r.json()
        for c in obj["connections"]:
            [x, y], host = c
            if x == 0 and y == 0:
                return host
        return None

    @overrides(SpallocJob.get_connections)
    def get_connections(self) -> Dict[XY, str]:
        r = self._get(self.__machine_url)
        if r.status_code == 204:
            return {}
        return {
            (int(x), int(y)): str(host)
            for ((x, y), host) in r.json()["connections"]
        }

    @property
    def __proxy_url(self) -> Optional[str]:
        """
        The URL for talking to the proxy connection system.
        """
        r = self._get(self._url)
        if r.status_code == 204:
            return None
        try:
            url = r.json()["proxy-ref"]
            logger.info("Connecting to proxy on {}", url)
            return url
        except KeyError:
            return None

    def __init_proxy(self) -> Tuple[_ProxyReceiver, WebSocket]:
        if self.__proxy_handle is None or not self.__proxy_handle.connected:
            if self.__proxy_url is None:
                raise ValueError("no proxy available")
            self.__proxy_handle = self._websocket(
                self.__proxy_url, origin=get_hostname(self._url))
            self.__proxy_thread = _ProxyReceiver(self.__proxy_handle)
            self.__proxy_ping = _ProxyPing(self.__proxy_handle)
        assert self.__proxy_handle is not None
        assert self.__proxy_thread is not None
        return self.__proxy_thread, self.__proxy_handle

    @overrides(SpallocJob.connect_to_board)
    def connect_to_board(
            self, x: int, y: int,
            port: int = SCP_SCAMP_PORT) -> SpallocSCPConnection:
        proxy, ws = self.__init_proxy()
        return _ProxiedSCAMPConnection(ws, proxy, int(x), int(y), int(port))

    @overrides(SpallocJob.connect_for_booting)
    def connect_for_booting(self) -> SpallocBootConnection:
        proxy, ws = self.__init_proxy()
        return _ProxiedBootConnection(ws, proxy)

    @overrides(SpallocJob.open_eieio_connection)
    def open_eieio_connection(self, x: int, y: int) -> SpallocEIEIOConnection:
        proxy, ws = self.__init_proxy()
        return _ProxiedEIEIOConnection(
            ws, proxy, int(x), int(y), SCP_SCAMP_PORT)

    @overrides(SpallocJob.open_eieio_listener_connection)
    def open_eieio_listener_connection(self) -> SpallocEIEIOListener:
        proxy, ws = self.__init_proxy()
        return _ProxiedEIEIOListener(ws, proxy, self.get_connections())

    @overrides(SpallocJob.open_udp_listener_connection)
    def open_udp_listener_connection(self) -> UDPConnection:
        proxy, ws = self.__init_proxy()
        return _ProxiedUDPListener(ws, proxy, self.get_connections())

    @overrides(SpallocJob.wait_for_state_change)
    def wait_for_state_change(self, old_state: SpallocState,
                              timeout: Optional[int] = None) -> SpallocState:
        while old_state != SpallocState.DESTROYED:
            obj = self._get(self._url, wait="true", timeout=timeout).json()
            s = SpallocState[obj["state"]]
            if s != old_state or s == SpallocState.DESTROYED:
                return s
        return old_state

    @overrides(SpallocJob.wait_until_ready)
    def wait_until_ready(self, timeout: Optional[int] = None,
                         n_retries: Optional[int] = None):
        state = self.get_state()
        retries = 0
        while (state != SpallocState.READY and
               (n_retries is None or retries < n_retries)):
            retries += 1
            state = self.wait_for_state_change(state, timeout=timeout)
            if state == SpallocState.DESTROYED:
                raise SpallocException("job was unexpectedly destroyed")

    @overrides(SpallocJob.destroy)
    def destroy(self, reason: str = "finished"):
        self._keepalive_url = None
        if self.__proxy_handle is not None:
            if self.__proxy_thread:
                self.__proxy_thread.close()
            if self.__proxy_ping:
                self.__proxy_ping.close()
            self.__proxy_handle.close()
        self._delete(self._url, reason=str(reason))
        logger.info("deleted job at {}", self._url)

    def __keepalive(self) -> bool:
        """
        Signal spalloc that we want the job to stay alive for a while longer.

        :return: False if the job has not been destroyed
        :rtype: bool
        """
        if self._keepalive_url is None:
            return False
        cookies, headers = self._session_credentials
        headers["Content-Type"] = "text/plain; charset=UTF-8"
        logger.debug(self._keepalive_url)
        requests.put(self._keepalive_url, data="alive", cookies=cookies,
                     headers=headers, allow_redirects=False, timeout=10)
        return True

    def __start_keepalive(self) -> None:
        """
        Method for keep alive thread to start the keep alive class

        """
        try:
            while self.__keepalive():
                time.sleep(KEEP_ALIVE_PERIOND / 2)
        except Exception as ex:  # pylint: disable=broad-except
            logger.exception(ex)

    @overrides(SpallocJob.where_is_machine)
    def where_is_machine(self, x: int, y: int) -> Optional[
            Tuple[int, int, int]]:
        r = self._get(self.__chip_url, x=int(x), y=int(y))
        if r.status_code == 204:
            return None
        return cast(Tuple[int, int, int], tuple(
            r.json()["physical-board-coordinates"]))

    @overrides(SpallocJob.create_transceiver)
    def create_transceiver(self) -> Transceiver:
        if self.get_state() != SpallocState.READY:
            raise SpallocException("job not ready to execute scripts")
        proxies: List[Connection] = [
            self.connect_to_board(x, y) for (x, y) in self.get_connections()]
        # Also need a boot connection
        proxies.append(self.connect_for_booting())
        return create_transceiver_from_connections(connections=proxies)

    def __repr__(self):
        return f"SpallocJob({self._url})"


class _ProxiedConnection(metaclass=AbstractBase):
    """
    Core multiplexer/demultiplexer emulating a connection that is proxied
    over a websocket.

    None of the methods are public because subclasses may expose a profile of
    them to conform to a particular type of connection.
    """

    def __init__(self, ws: WebSocket, receiver: _ProxyReceiver):
        self.__ws: Optional[WebSocket] = ws
        self.__receiver: Optional[_ProxyReceiver] = receiver
        self.__msgs: queue.SimpleQueue = queue.SimpleQueue()
        self.__call_queue: queue.Queue = queue.Queue(1)
        self.__call_lock = threading.RLock()
        self.__current_msg: Optional[bytes] = None
        self.__handle = self._open_connection()
        self.__receiver.listen(self.__handle, self.__msgs.put)

    @abstractmethod
    def _open_connection(self) -> int:
        raise NotImplementedError

    def _call(self, protocol: ProxyProtocol, packer: struct.Struct,
              unpacker: struct.Struct, *args) -> Tuple[Any, ...]:
        """
        Do a synchronous call.

        :param protocol:
            The protocol message number.
        :param packer:
            How to form the protocol message. The first two arguments passed
            will be the protocol message number and an issued correlation ID
            (not needed by the caller).
        :param unpacker:
            How to extract the expected response.
        :param args:
            Additional arguments to pass to the packer.
        :return:
            The results from the unpacker *after* the protocol message code and
            the correlation ID.
        :raises IOError:
            If something goes wrong. This could be due to the websocket being
            closed, or the receipt of an ERROR response.
        """
        if not self._connected:
            raise IOError("socket closed")
        if not self.__receiver:
            raise IOError("socket closed")
        if not self.__ws:
            raise IOError("socket closed")
        with self.__call_lock:
            # All calls via websocket use correlation_id
            correlation_id = self.__receiver.expect_return(
                self.__call_queue.put)
            self.__ws.send_binary(packer.pack(protocol, correlation_id, *args))
            if not self._connected:
                raise IOError("socket closed after send!")
            reply = self.__call_queue.get()
            code, _ = _msg.unpack_from(reply, 0)
            if code == ProxyProtocol.ERROR:
                # Rest of message is UTF-8 encoded error message string
                payload = reply[_msg.size:].decode("utf-8")
                if len(payload):
                    raise _ProxyServiceError(payload)
                raise _ProxyServiceError(
                    f"unknown problem with {protocol} call")
            return unpacker.unpack(reply)[2:]

    @property
    def _connected(self) -> bool:
        return bool(self.__ws and self.__ws.connected)

    def _throw_if_closed(self) -> None:
        if not self._connected:
            raise IOError("socket closed")

    def _close(self) -> None:
        if self._connected:
            channel_id, = self._call(
                ProxyProtocol.CLOSE, _close_req, _open_close_res,
                self.__handle)
            if channel_id != self.__handle:
                raise IOError("failed to close proxy socket")
        if self.__receiver:
            self.__receiver.unlisten(self.__handle)
        self.__ws = None
        self.__receiver = None

    def _send(self, message: bytes):
        self._throw_if_closed()
        # Put the header on the front and send it
        if not self.__ws:
            raise IOError("socket closed")
        self.__ws.send_binary(_msg.pack(
            ProxyProtocol.MSG, self.__handle) + message)

    def _send_to(self, message: bytes, x: int, y: int, port: int):
        self._throw_if_closed()
        # Put the header on the front and send it
        if not self.__ws:
            raise IOError("socket closed")
        self.__ws.send_binary(_msg_to.pack(
            ProxyProtocol.MSG_TO, self.__handle, x, y, port) + message)

    def __get(self, timeout: float = 0.5) -> bytes:
        """
        Get a value from the queue. Handles block/non-block switching and
        trimming of the message protocol prefix.
        """
        if not timeout:
            return self.__msgs.get(block=False)[_msg.size:]
        else:
            return self.__msgs.get(timeout=timeout)[_msg.size:]

    def _receive(self, timeout: Optional[float] = None) -> bytes:
        if self.__current_msg is not None:
            try:
                return self.__current_msg
            finally:
                self.__current_msg = None
        if timeout is None:
            while True:
                try:
                    return self.__get()
                except queue.Empty:
                    self._throw_if_closed()
        else:
            try:
                return self.__get(timeout)
            except queue.Empty as e:
                self._throw_if_closed()
                raise SpinnmanTimeoutException("receive", timeout) from e

    def _is_ready_to_receive(self, timeout: float = 0) -> bool:
        # If we already have a message or the queue peek succeeds, return now
        if self.__current_msg is not None or not self.__msgs.empty():
            return True
        try:
            self.__current_msg = self.__get(timeout)
            return True
        except queue.Empty:
            return False


class _ProxiedBidirectionalConnection(
        _ProxiedConnection, SpallocProxiedConnection):
    """
    A connection that talks to a particular board via the proxy.
    """

    def __init__(
            self, ws: WebSocket, receiver: _ProxyReceiver,
            x: int, y: int, port: int):
        self.__connect_args = (x, y, port)
        super().__init__(ws, receiver)

    @overrides(_ProxiedConnection._open_connection)
    def _open_connection(self) -> int:
        handle, = self._call(
            ProxyProtocol.OPEN, _open_req, _open_close_res,
            *self.__connect_args)
        return handle

    @overrides(Connection.is_connected)
    def is_connected(self) -> bool:
        """
        Determines if the medium is connected at this point in time.

        :return: True if the medium is connected, False otherwise
        :rtype: bool
        """
        return self._connected

    @overrides(Connection.close)
    def close(self) -> None:
        """
        Closes the connection.
        """
        self._close()

    @overrides(SpallocProxiedConnection.send)
    def send(self, data: bytes):
        if not isinstance(data, (bytes, bytearray)):
            data = bytes(data)
        self._send(data)

    @overrides(SpallocProxiedConnection.receive)
    def receive(self, timeout: Optional[float] = None) -> bytes:
        return self._receive(timeout)

    @overrides(Listenable.is_ready_to_receive)
    def is_ready_to_receive(self, timeout: float = 0) -> bool:
        return self._is_ready_to_receive(timeout)

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError


class _ProxiedUnboundConnection(
        _ProxiedConnection, SpallocProxiedConnection):
    """
    A connection that can listen to all boards via the proxy, but which can
    only send if a target board is provided.
    """

    def __init__(self, ws: WebSocket, receiver: _ProxyReceiver):
        super().__init__(ws, receiver)
        self.__addr: Optional[str] = None
        self.__port: Optional[int] = None

    @overrides(_ProxiedConnection._open_connection)
    def _open_connection(self) -> int:
        handle, ip1, ip2, ip3, ip4, self.__port = self._call(
            ProxyProtocol.OPEN_UNBOUND, _open_listen_req, _open_listen_res)
        # Assemble the address into the format expected elsewhere
        self.__addr = f"{ip1}.{ip2}.{ip3}.{ip4}"
        return handle

    @property
    def _addr(self) -> Optional[str]:
        return self.__addr if self._connected else None

    @property
    def _port(self) -> Optional[int]:
        return self.__port if self._connected else None

    @overrides(Connection.is_connected)
    def is_connected(self) -> bool:
        """
        Determines if the medium is connected at this point in time.

        :return: True if the medium is connected, False otherwise
        :rtype: bool
        """
        return self._connected

    @overrides(Connection.close)
    def close(self) -> None:
        """
        Closes the connection.
        """
        self._close()

    @overrides(SpallocProxiedConnection.send)
    def send(self, data: bytes):
        self._throw_if_closed()
        raise IOError("socket is not open for sending")

    @overrides(SpallocProxiedConnection.receive)
    def receive(self, timeout: Optional[float] = None) -> bytes:
        return self._receive(timeout)

    @overrides(Listenable.is_ready_to_receive)
    def is_ready_to_receive(self, timeout: float = 0) -> bool:
        return self._is_ready_to_receive(timeout)

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError


class _ProxiedSCAMPConnection(
        _ProxiedBidirectionalConnection, SpallocSCPConnection):
    __slots__ = ("__chip_x", "__chip_y")

    def __init__(
            self, ws: WebSocket, receiver: _ProxyReceiver,
            x: int, y: int, port: int):
        super().__init__(ws, receiver, x, y, port)
        SpallocSCPConnection.__init__(self, x, y)

    def __str__(self):
        return f"SCAMPConnection[proxied]({self.chip_x},{self.chip_y})"


class _ProxiedBootConnection(
        _ProxiedBidirectionalConnection, SpallocBootConnection):
    __slots__ = ()

    def __init__(self, ws: WebSocket, receiver: _ProxyReceiver):
        super().__init__(ws, receiver, 0, 0, UDP_BOOT_CONNECTION_DEFAULT_PORT)

    def __str__(self):
        return "BootConnection[proxied]()"


class _ProxiedEIEIOConnection(
        _ProxiedBidirectionalConnection,
        SpallocEIEIOConnection, SpallocProxiedConnection):
    # Special: This is a unidirectional receive-only connection
    __slots__ = ("__addr", "__port", "__chip_x", "__chip_y")

    def __init__(
            self, ws: WebSocket, receiver: _ProxyReceiver,
            x: int, y: int, port: int):
        super().__init__(ws, receiver, x, y, port)
        self.__chip_x = x
        self.__chip_y = y

    @property
    @overrides(SpallocEIEIOConnection._coords)
    def _coords(self) -> XY:
        return self.__chip_x, self.__chip_y

    def send_to(
            self,
            data: bytes, address: tuple):  # pylint: disable=unused-argument
        """
        Direct ``send_to`` is unsupported.
        """
        self._throw_if_closed()
        raise IOError("socket is not open for sending")

    def __str__(self):
        return (f"EIEIOConnection[proxied](remote:{self.__chip_x},"
                f"{self.__chip_y})")


class _ProxiedEIEIOListener(_ProxiedUnboundConnection, SpallocEIEIOListener):
    __slots__ = ("__conns", )

    def __init__(self, ws: WebSocket, receiver: _ProxyReceiver,
                 conns: Dict[XY, str]):
        super().__init__(ws, receiver)
        # Invert the map
        self.__conns = {ip: xy for (xy, ip) in conns.items()}

    @overrides(SpallocEIEIOListener.send_to_chip)
    def send_to_chip(
            self, message: bytes, x: int, y: int, port: int = SCP_SCAMP_PORT):
        if not isinstance(message, (bytes, bytearray)):
            message = bytes(message)
        self._send_to(bytes(message), x, y, port)

    @property
    @overrides(SpallocEIEIOListener.local_ip_address)
    def local_ip_address(self) -> str:
        return self._addr or "0.0.0.0"

    @property
    @overrides(SpallocEIEIOListener.local_port)
    def local_port(self) -> int:
        return self._port or 0

    @overrides(SpallocEIEIOListener._get_chip_coords)
    def _get_chip_coords(self, ip_address: str) -> XY:
        return self.__conns[str(ip_address)]

    def __str__(self):
        return f"EIEIOConnection[proxied](local:{self._addr}:{self._port})"


class _ProxiedUDPListener(_ProxiedUnboundConnection, UDPConnection):
    __slots__ = ("__conns", )

    def __init__(self, ws: WebSocket, receiver: _ProxyReceiver,
                 conns: Dict[XY, str]):
        super().__init__(ws, receiver)
        # Invert the map
        self.__conns = {ip: xy for (xy, ip) in conns.items()}

    @overrides(UDPConnection.send_to)
    def send_to(self, data: bytes, address: Tuple[str, int]):
        ip, port = address
        x, y = self.__conns[ip]
        self._send_to(data, x, y, port)

    @property
    @overrides(UDPConnection.local_ip_address)
    def local_ip_address(self) -> str:
        return self._addr or "0.0.0.0"

    @property
    @overrides(UDPConnection.local_port)
    def local_port(self) -> int:
        return self._port or 0

    def __str__(self):
        return f"UDPConnection[proxied](local:{self._addr}:{self._port})"



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from spinnman.constants import EIEIO_COMMAND_IDS
from .eieio_command_message import EIEIOCommandMessage
from .eieio_command_header import EIEIOCommandHeader


class NotificationProtocolDatabaseLocation(EIEIOCommandMessage):
    """
    Packet which contains the path to the database created by the
    toolchain which is to be used by any software which interfaces with
    SpiNNaker. Also the acknowledgement of that message.

    This message is not sent to SpiNNaker boards but rather to an auxiliary
    tool (e.g., data visualiser).
    """
    __slots__ = "_database_path",

    def __init__(self, database_path=None):
        """
        :param str database_path:
            The location of the database. If ``None``, this is an
            acknowledgement, stating that the database has now been read.
        """
        super().__init__(EIEIOCommandHeader(EIEIO_COMMAND_IDS.DATABASE))
        self._database_path = None
        if database_path is not None:
            self._database_path = database_path.encode()

    @property
    def database_path(self) -> Optional[str]:
        """
        Gets the database path passed into the init.

        The path is encoded by the init and decode back to a str here.

        :rtype: str
        """
        if self._database_path is not None:
            return self._database_path.decode()
        else:
            return None

    @property
    def bytestring(self):
        data = super().bytestring
        if self._database_path is not None:
            data += self._database_path
        return data

    @staticmethod
    def from_bytestring(command_header, data, offset):
        database_path = None
        if len(data) - offset > 0:
            database_path = data[offset:]
        return NotificationProtocolDatabaseLocation(database_path)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Callable

from spinn_utilities.typing.coords import XYP

from spinnman.messages.scp.impl import ReadLink, ReadMemory
from spinnman.messages.scp.impl.read_memory import Response
from spinnman.constants import UDP_MESSAGE_MAX_SIZE
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest

from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)


class ReadMemoryProcess(AbstractMultiConnectionProcess[Response]):
    """
    A process for reading memory on a SpiNNaker chip.
    """
    __slots__ = ("_view", )

    def __init__(self, connection_selector: ConnectionSelector):
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector)
        self._view = memoryview(b'')

    def __handle_response(self, offset: int, response: Response):
        self._view[offset:offset + response.length] = response.data[
            response.offset:response.offset + response.length]

    def read_memory(self, coordinates: XYP, base_address: int,
                    length: int) -> bytearray:
        """
        Read some memory from a core.

        :param tuple(int,int,int) coordinates:
        :param int base_address:
        :param int length:
        :rtype: bytearray
        """
        return self._read_memory(
            base_address, length,
            functools.partial(ReadMemory, coordinates))

    def read_link_memory(self, coordinates: XYP, link: int,
                         base_address: int, length: int) -> bytearray:
        """
        Read some memory from the neighbour of a core.

        :param tuple(int,int,int) coordinates:
        :param int link:
        :param int base_address:
        :param int length:
        :rtype: bytearray
        """
        return self._read_memory(
            base_address, length,
            functools.partial(ReadLink, coordinates, link))

    def _read_memory(
            self, base_address: int, length: int,
            packet_class: Callable[
                [int, int], AbstractSCPRequest[Response]]) -> bytearray:
        data = bytearray(length)
        self._view = memoryview(data)
        n_bytes = length
        offset = 0

        with self._collect_responses():
            while n_bytes > 0:
                bytes_to_get = min((n_bytes, UDP_MESSAGE_MAX_SIZE))
                self._send_request(
                    packet_class(base_address + offset, bytes_to_get),
                    functools.partial(self.__handle_response, offset))
                n_bytes -= bytes_to_get
                offset += bytes_to_get

        return data



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .get_chip_info_response import GetChipInfoResponse


class GetChipInfo(AbstractSCPRequest[GetChipInfoResponse]):
    """
    An SCP request to read the chip information from a core.
    """
    __slots__ = ()

    def __init__(self, x: int, y: int, with_size: bool = False):
        """
        :param int x:
            The x-coordinate of the chip to read from, between 0 and 255
        :param int y:
            The y-coordinate of the chip to read from, between 0 and 255
        :param bool with_size:
            Whether the size should be included in the response
        """
        # Bits 0-4 + bit 6 = all information except size
        argument_1 = 0x5F
        if with_size:

            # Bits 0-6 = all information including size
            argument_1 = 0x7F

        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_INFO),
            argument_1=argument_1)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> GetChipInfoResponse:
        return GetChipInfoResponse()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from contextlib import suppress
from typing import Optional
from spinnman.constants import UDP_BOOT_CONNECTION_DEFAULT_PORT
from .udp_connection import UDPConnection

_BOOTROM_SPINN_PORT = 54321  # Matches SPINN_PORT in spinnaker_bootROM


class IPAddressesConnection(UDPConnection):
    """
    A connection that detects any UDP packet that is transmitted by
    SpiNNaker boards prior to boot.
    """
    __slots__ = ()

    def __init__(self, local_host=None,
                 local_port=UDP_BOOT_CONNECTION_DEFAULT_PORT):
        super().__init__(local_host=local_host, local_port=local_port)

    def receive_ip_address(self, timeout:  Optional[float] = None
                           ) -> Optional[str]:
        """

        :param timeout:
        :type timeout: float or None
        :rtype: str or None
        """
        with suppress(Exception):
            (_, (ip_address, port)) = self.receive_with_address(timeout)
            if port == _BOOTROM_SPINN_PORT:
                return ip_address
        return None

    def __repr__(self):
        return f"IPAddressesConnection(local_host={self.local_ip_address}," \
               f" local_port={self.local_port})"



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union
import struct

from spinn_utilities.overrides import overrides
from spinnman.exceptions import (
    SpinnmanInvalidPacketException, SpinnmanInvalidParameterTypeException)
from spinnman.constants import EIEIO_COMMAND_IDS

from .eieio_command_message import EIEIOCommandMessage
from .eieio_command_header import EIEIOCommandHeader

_PATTERN_BBBB = struct.Struct("<BBBB")
_PATTERN_BBII = struct.Struct("<BBII")
_PATTERN_xxBBII = struct.Struct("<xxBBII")
_PATTERN_BB = struct.Struct("<BB")
_PATTERN_BBBBII = struct.Struct("<BBBBII")


class SpinnakerRequestReadData(EIEIOCommandMessage):
    """
    Message used in the context of the buffering output mechanism which is
    sent from the SpiNNaker system to the host computer to signal that
    some data is available to be read.
    """
    __slots__ = (
        "_header",
        "_requests")

    def __init__(
            self, x: int, y: int, p: int, region_id: Union[List[int], int],
            sequence_no: int, n_requests: int, channel: Union[List[int], int],
            start_address: Union[List[int], int],
            space_to_be_read: Union[List[int], int]):
        """

        :param int x:
        :param int y:
        :param int p:
        :param region_id:
        :type region_id: list(int) or int
        :param int sequence_no:
        :param int n_requests:
        :param channel:
        :type channel: list(int) or int
        :param start_address:
        :type start_address:: list(int) or int
        :param space_to_be_read:
        :type space_to_be_read: list(int) or int
        """
        # pylint: disable=too-many-arguments
        if not isinstance(channel, list):
            channel = [channel]

        if not isinstance(region_id, list):
            region_id = [region_id]

        if not isinstance(start_address, list):
            start_address = [start_address]

        if not isinstance(space_to_be_read, list):
            space_to_be_read = [space_to_be_read]

        if len(start_address) != n_requests or \
                len(space_to_be_read) != n_requests or \
                len(channel) != n_requests or \
                len(region_id) != n_requests:
            raise SpinnmanInvalidPacketException(
                "SpinnakerRequestReadData",
                "The format for a SpinnakerRequestReadData packet is "
                f"invalid: {n_requests:d} request(s), "
                f"{len(start_address):d} start address(es) defined, "
                f"{len(space_to_be_read):d} space(s) to be read defined, "
                f"{len(region_id):d} region(s) defined, "
                f"{len(channel):d} channel(s) defined")

        super().__init__(EIEIOCommandHeader(
            EIEIO_COMMAND_IDS.SPINNAKER_REQUEST_READ_DATA))
        self._header = _SpinnakerRequestReadDataHeader(
            x, y, p, n_requests, sequence_no)
        self._requests = _SpinnakerRequestReadDataRequest(
            channel, region_id, start_address, space_to_be_read)

    @property
    def x(self) -> int:
        """
        The x value passed into the init.

        :rtype: int
        """
        return self._header.x

    @property
    def y(self) -> int:
        """
        The y value passed into the init.

        :rtype: int
        """
        return self._header.y

    @property
    def p(self) -> int:
        """
        The p value passed into the init.

        :rtype: int
        """
        return self._header.p

    @property
    def n_requests(self) -> int:
        """
        The n_requests value passed into the init.

        :rtype: int
        """
        return self._header.n_requests

    @property
    def sequence_no(self) -> int:
        """
        The sequence_no value passed into the init.

        :rtype: int
        """
        return self._header.sequence_no

    def channel(self, request_id: int) -> int:
        """
        The channel for this request_id.

        :param int request_id:
        :rtype: int
        :raises IndexError: If the request_id is invalid
        """
        return self._requests.channel(request_id)

    def region_id(self, request_id) -> int:
        """
        The region_id for this request_id.

        :param int request_id:
        :rtype: int
        :raises IndexError: If the request_id is invalid
        """
        return self._requests.region_id(request_id)

    def start_address(self, request_id) -> int:
        """
        The start_address for this request_id.

        :param int request_id:
        :rtype: int
        :raises IndexError: If the request_id is invalid
        """
        return self._requests.start_address(request_id)

    def space_to_be_read(self, request_id) -> int:
        """
        The space_to_be_read for this request_id.

        :param int request_id:
        :rtype: int
        :raises IndexError: If the request_id is invalid
        """
        return self._requests.space_to_be_read(request_id)

    @staticmethod
    @overrides(EIEIOCommandMessage.get_min_packet_length)
    def get_min_packet_length() -> int:
        return 16

    @staticmethod
    @overrides(EIEIOCommandMessage.from_bytestring)
    def from_bytestring(command_header: EIEIOCommandHeader, data: bytes,
                        offset: int) -> "SpinnakerRequestReadData":
        (y, x, processor_and_requests, sequence_no) = \
            _PATTERN_BBBB.unpack_from(data, offset)
        p = (processor_and_requests >> 3) & 0x1F
        n_requests = processor_and_requests & 0x7
        offset += 4
        channel = list()
        region_id = list()
        start_address = list()
        space_to_be_read = list()

        for i in range(n_requests):
            if i == 0:
                request_channel, request_region_id, request_start_address, \
                    request_space_to_be_read = _PATTERN_BBII.unpack_from(
                        data, offset)
                offset += 10
            else:
                request_channel, request_region_id, request_start_address, \
                    request_space_to_be_read = _PATTERN_xxBBII.unpack_from(
                        data, offset)
                offset += 12
            channel.append(request_channel)
            region_id.append(request_region_id)
            start_address.append(request_start_address)
            space_to_be_read.append(request_space_to_be_read)
        return SpinnakerRequestReadData(
            x, y, p, region_id, sequence_no, n_requests, channel,
            start_address, space_to_be_read)

    @property
    @overrides(EIEIOCommandMessage.bytestring)
    def bytestring(self) -> bytes:
        byte_string = super().bytestring
        byte_string += _PATTERN_BB.pack(self.x, self.y)
        n_requests = self.n_requests
        processor_and_request = (self.p << 3) | n_requests

        for i in range(n_requests):
            if i == 0:
                byte_string += _PATTERN_BBBBII.pack(
                    processor_and_request, self.sequence_no,
                    self.channel(i), self.region_id(i),
                    self.start_address(0), self.space_to_be_read(i))
            else:
                byte_string += _PATTERN_xxBBII.pack(
                    self.channel(i), self.region_id(i),
                    self.start_address(0), self.space_to_be_read(i))
        return byte_string


class _SpinnakerRequestReadDataHeader(object):
    """
    Contains the position of the core in the machine (x, y, p), the number
    of requests and a sequence number.
    """
    __slots__ = [
        "_n_requests",
        "_sequence_no",
        "_p", "_x", "_y"]

    def __init__(
            self, x: int, y: int, p: int, n_requests: int, sequence_no: int):
        """

        :param int x:
        :param int y:
        :param int p:
        :param int n_requests:
        :param int sequence_no:
        """
        # pylint: disable=too-many-arguments
        self._x = x
        self._y = y
        self._p = p
        self._n_requests = n_requests
        self._sequence_no = sequence_no

    @property
    def x(self) -> int:
        """
        The x value passed into the init.

        :rtype: int
        """
        return self._x

    @property
    def y(self) -> int:
        """
        The y value passed into the init.

        :rtype: int
        """
        return self._y

    @property
    def p(self) -> int:
        """
        The p value passed into the init.

        :rtype: int
        """
        return self._p

    @property
    def sequence_no(self) -> int:
        """
        The sequence_no value passed into the init.

        :rtype: int
        """
        return self._sequence_no

    @property
    def n_requests(self) -> int:
        """
        The n_request value passed into the init.

        :rtype: int
        """
        return self._n_requests


class _SpinnakerRequestReadDataRequest(object):
    """
    Contains a set of requests which refer to the channels used.
    """
    __slots__ = [
        "_channel",
        "_region_id",
        "_start_address",
        "_space_to_be_read"]

    def __init__(self, channel: Union[List[int], int],
                 region_id: Union[List[int], int],
                 start_address: Union[List[int], int],
                 space_to_be_read: Union[List[int], int]):
        """

        :param channel:
        :type channel: list(int) or int
        :param region_id:
        :type region_id: list(int) or int
        :param start_address:
        :type start_address: list(int) or int
        :param space_to_be_read:
        :type space_to_be_read: list(int) or int
        """
        if not isinstance(channel, list):
            self._channel = [channel]
        else:
            self._channel = channel

        if not isinstance(region_id, list):
            self._region_id = [region_id]
        else:
            self._region_id = region_id

        if not isinstance(start_address, list):
            self._start_address = [start_address]
        else:
            self._start_address = start_address

        if not isinstance(space_to_be_read, list):
            self._space_to_be_read = [space_to_be_read]
        else:
            self._space_to_be_read = space_to_be_read

    def channel(self, request_id) -> int:
        """
        Gets the channel for this request_id

        :param int request_id:
        :rtype: int
        :raises SpinnmanInvalidParameterTypeException:
            if the request_id os too high
        """

        if len(self._channel) > request_id:
            return self._channel[request_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer",
            f"channel request needs to be comprised between 0 and "
            f"{len(self._channel) - 1:d}; current value: {request_id:d}")

    def region_id(self, request_id) -> int:
        """
        Gets the region_id for this request_id

        :param int request_id:
        :rtype: int
        :raises SpinnmanInvalidParameterTypeException:
            if the request_id os too high
        """
        if len(self._region_id) > request_id:
            return self._region_id[request_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer",
            f"region ID request needs to be comprised between 0 and "
            f"{len(self._region_id) - 1:d}; current value: {request_id:d}")

    def start_address(self, request_id: int) -> int:
        """
        Gets the start address for this request_id

        :param int request_id:
        :rtype: int
        :raises SpinnmanInvalidParameterTypeException:
            if the request_id os too high
        """
        if len(self._start_address) > request_id:
            return self._start_address[request_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer",
            f"start address request needs to be comprised between 0 and "
            f"{len(self._start_address) - 1:d}; current value: {request_id:d}")

    def space_to_be_read(self, request_id: int) -> int:
        """
        Checks if there is enough space to request this id

        :rtype: bool
        :raises SpinnmanInvalidParameterTypeException:
            if the request_id os too high
        """
        if len(self._space_to_be_read) > request_id:
            return self._space_to_be_read[request_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer",
            f"space to be read request needs to be comprised between 0 and "
            f"{len(self._space_to_be_read) - 1:d}; "
            f"current value: {request_id:d}")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Sequence
from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.constants import ROUTER_REGISTER_REGISTERS
from spinnman.model.enums.router_error import RouterError


class RouterDiagnostics(object):
    """
    Represents a set of diagnostic information available from a chip router.
    """
    __slots__ = [
        "_error_status",
        "_mon",
        "_register_values",
        "_wait_1",
        "_wait_2"]

    def __init__(
            self, control_register: int, error_status: int,
            register_values: Sequence[int]):
        """
        :param int control_register: The value of the control register
        :param int error_status: The value of the error_status
        :param list(int) register_values:
            The values of the 16 router registers
        :raise SpinnmanInvalidParameterException:
            If the number of register values is not 16
        """
        if len(register_values) != 16:
            raise SpinnmanInvalidParameterException(
                "len(register_values)",
                str(len(ROUTER_REGISTER_REGISTERS)),
                "There must be exactly 16 register values")

        self._mon = (control_register >> 8) & 0x1F
        self._wait_1 = (control_register >> 16) & 0xFF
        self._wait_2 = (control_register >> 24) & 0xFF

        self._error_status = error_status

        self._register_values = register_values

    # pylint: disable=wrong-spelling-in-docstring
    @property
    def mon(self) -> int:
        """
        The "mon" part of the control register.

        :rtype: int
        """
        return self._mon

    @property
    def wait_1(self) -> int:
        """
        The "wait_1" part of the control register.

        :rtype: int
        """
        return self._wait_1

    @property
    def wait_2(self) -> int:
        """
        The "wait_2" part of the control register.

        :rtype: int
        """
        return self._wait_2

    @property
    def error_status(self) -> int:
        """
        The error status.

        :rtype: int
        """
        return self._error_status

    @property
    def error_count(self) -> int:
        """
        The count of errors.

        :rtype: int
        """
        return self._error_status & 0xFF

    @property
    def errors_set(self) -> Sequence[RouterError]:
        """
        A list of errors that have been detected.

        :rtype: list(RouterError)
        """
        return [
            error for error in RouterError if error.value & self._error_status]

    @property
    def n_local_multicast_packets(self) -> int:
        """
        The number of multicast packets received from local cores.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.LOC_MC.value]

    @property
    def n_external_multicast_packets(self) -> int:
        """
        The number of multicast packets received from external links.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.EXT_MC.value]

    @property
    def n_dropped_multicast_packets(self) -> int:
        """
        The number of multicast packets received that were dropped.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.DUMP_MC.value]

    @property
    def n_local_peer_to_peer_packets(self) -> int:
        """
        The number of peer-to-peer packets received from local cores.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.LOC_PP.value]

    @property
    def n_external_peer_to_peer_packets(self) -> int:
        """
        The number of peer-to-peer packets received from external links.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.EXT_PP.value]

    @property
    def n_dropped_peer_to_peer_packets(self) -> int:
        """
        The number of peer-to-peer packets received that were dropped.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.DUMP_PP.value]

    @property
    def n_local_nearest_neighbour_packets(self) -> int:
        """
        The number of nearest-neighbour packets received from local cores.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.LOC_NN.value]

    @property
    def n_external_nearest_neighbour_packets(self) -> int:
        """
        The number of nearest-neighbour packets received from external links.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.EXT_NN.value]

    @property
    def n_dropped_nearest_neighbour_packets(self) -> int:
        """
        The number of nearest-neighbour packets received that were dropped.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.DUMP_NN.value]

    @property
    def n_local_fixed_route_packets(self) -> int:
        """
        The number of fixed-route packets received from local cores.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.LOC_FR.value]

    @property
    def n_external_fixed_route_packets(self) -> int:
        """
        The number of fixed-route packets received from external links.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.EXT_FR.value]

    @property
    def n_dropped_fixed_route_packets(self) -> int:
        """
        The number of fixed-route packets received that were dropped.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.DUMP_FR.value]

    @property
    def user_0(self) -> int:
        """
        The number of packets counted by the user 0 router diagnostic filter.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.USER_0.value]

    @property
    def user_1(self) -> int:
        """
        The number of packets counted by the user 1 router diagnostic filter.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.USER_1.value]

    @property
    def user_2(self) -> int:
        """
        The number of packets counted by the user 2 router diagnostic filter.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.USER_2.value]

    @property
    def user_3(self) -> int:
        """
        The number of packets counted by the user 3 router diagnostic filter.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.USER_3.value]

    @property
    def user_registers(self) -> Sequence[int]:
        """
        The values in the user control registers.

        :return: An array of 4 values
        :rtype: list(int)
        """
        return self._register_values[
            ROUTER_REGISTER_REGISTERS.USER_0.value:
            ROUTER_REGISTER_REGISTERS.USER_3.value + 1]

    @property
    def registers(self) -> Sequence[int]:
        """
        The values in all of the registers.  Can be used to directly access
        the registers if they have been programmed to give different values.

        :return: An array of 16 values
        :rtype: array(int)
        """
        return self._register_values



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
import functools
import struct
from collections import defaultdict
from typing import Dict, Iterable, List
from spinn_utilities.typing.coords import XYP
from spinn_machine import CoreSubsets
from spinnman.model import IOBuffer
from spinnman.utilities.utility_functions import get_vcpu_address
from spinnman.messages.scp.impl.read_memory import ReadMemory, Response
from spinnman.constants import UDP_MESSAGE_MAX_SIZE, CPU_IOBUF_ADDRESS_OFFSET
from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)


@dataclass(frozen=True)
class _RegionTail:
    scamp_coords: XYP
    core_coords: XYP
    n: int
    base_address: int
    size: int
    offset: int


@dataclass(frozen=True)
class _NextRegion:
    scamp_coords: XYP
    core_coords: XYP
    n: int
    next_address: int
    first_read_size: int

    def next_at(self, address: int) -> '_NextRegion':
        """

        :param int address:
        :rtype: _NextRegion
        """
        return _NextRegion(
            self.scamp_coords, self.core_coords, self.n + 1, address,
            self.first_read_size)

    def tail(self, address: int, size: int, offset: int) -> _RegionTail:
        """

        :param int address:
        :param int size:
        :param int offset:
        :rtype: _RegionTail
        """
        return _RegionTail(
            self.scamp_coords, self.core_coords, self.n, address, size, offset)


_ENCODING = "ascii"
_ONE_WORD = struct.Struct("<I")
_FIRST_IOBUF = struct.Struct("<I8xI")


class ReadIOBufProcess(AbstractMultiConnectionProcess[Response]):
    """
    A process for reading IOBUF memory (mostly log messages) from a
    SpiNNaker core.
    """
    __slots__ = (
        "_extra_reads",
        "_iobuf",
        "_iobuf_address",
        "_iobuf_view",
        "_next_reads")

    def __init__(self, connection_selector: ConnectionSelector) -> None:
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector)

        # A dictionary of (x, y, p) -> iobuf address
        self._iobuf_address: Dict[XYP, int] = dict()

        # A dictionary of (x, y, p) -> OrderedDict(n) -> bytearray
        self._iobuf: Dict[XYP, Dict[int, bytes]] = defaultdict(dict)

        # A dictionary of (x, y, p) -> OrderedDict(n) -> memoryview
        self._iobuf_view: Dict[XYP, Dict[int, memoryview]] = defaultdict(dict)

        # A list of extra reads that need to be done as a result of the first
        # read = list of (x, y, p, n, base_address, size, offset)
        self._extra_reads: List[_RegionTail] = list()

        # A list of next reads that need to be done as a result of the first
        # read = list of (x, y, p, n, next_address, first_read_size)
        self._next_reads: List[_NextRegion] = list()

    def _request_iobuf_address(self, iobuf_size: int, x: int, y: int, p: int):
        scamp_coords = (x, y, 0)
        base_address = get_vcpu_address(p) + CPU_IOBUF_ADDRESS_OFFSET
        self._send_request(
            ReadMemory(scamp_coords, base_address, 4),
            functools.partial(self.__handle_iobuf_address_response,
                              iobuf_size, scamp_coords, (x, y, p)))

    def __handle_iobuf_address_response(
            self, iobuf_size: int, scamp_coords: XYP, xyp: XYP,
            response: Response):
        iobuf_address, = _ONE_WORD.unpack_from(response.data, response.offset)
        if iobuf_address != 0:
            first_read_size = min((iobuf_size + 16, UDP_MESSAGE_MAX_SIZE))
            self._next_reads.append(_NextRegion(
                scamp_coords, xyp, 0, iobuf_address, first_read_size))

    def _request_iobuf_region_tail(self, tail: _RegionTail):
        self._send_request(
            ReadMemory(tail.scamp_coords, tail.base_address, tail.size),
            functools.partial(
                self.__handle_extra_iobuf_response, tail))

    def __handle_extra_iobuf_response(
            self, tail: _RegionTail, response: Response):
        view = self._iobuf_view[tail.core_coords][tail.n]
        base = tail.offset
        view[base:base + response.length] = response.data[
            response.offset:response.offset + response.length]

    def _request_iobuf_region(self, region: _NextRegion):
        self._send_request(
            ReadMemory(region.scamp_coords, region.next_address,
                       region.first_read_size),
            functools.partial(self.__handle_first_iobuf_response, region))

    def __handle_first_iobuf_response(
            self, region: _NextRegion, response: Response):
        base_address = region.next_address

        # Unpack the iobuf header
        (next_address, bytes_to_read) = _FIRST_IOBUF.unpack_from(
            response.data, response.offset)

        # Create a buffer for the data
        data = bytearray(bytes_to_read)
        view = memoryview(data)
        self._iobuf[region.core_coords][region.n] = data
        self._iobuf_view[region.core_coords][region.n] = view

        # Put the data from this packet into the buffer
        packet_bytes = response.length - 16
        if packet_bytes > bytes_to_read:
            packet_bytes = bytes_to_read
        if packet_bytes > 0:
            offset = response.offset + 16
            view[0:packet_bytes] = response.data[offset:offset + packet_bytes]

        bytes_to_read -= packet_bytes
        base_address += packet_bytes + 16
        read_offset = packet_bytes

        # While more reads need to be done to read the data
        while bytes_to_read > 0:
            # Read the next bit of memory making up the buffer
            next_bytes_to_read = min((bytes_to_read, UDP_MESSAGE_MAX_SIZE))
            self._extra_reads.append(region.tail(
                base_address, next_bytes_to_read, read_offset))
            base_address += next_bytes_to_read
            read_offset += next_bytes_to_read
            bytes_to_read -= next_bytes_to_read

        # If there is another IOBuf buffer, read this next
        if next_address != 0:
            self._next_reads.append(region.next_at(next_address))

    def read_iobuf(
            self, iobuf_size: int,
            core_subsets: CoreSubsets) -> Iterable[IOBuffer]:
        """
        :param int iobuf_size:
        :param ~spinn_machine.CoreSubsets core_subsets:
        :rtype: iterable(IOBuffer)
        """
        # Get the iobuf address for each core
        with self._collect_responses():
            for core_subset in core_subsets:
                x, y = core_subset.x, core_subset.y
                for p in core_subset.processor_ids:
                    self._request_iobuf_address(iobuf_size, x, y, p)

        # Run rounds of the process until reading is complete
        while self._extra_reads or self._next_reads:
            with self._collect_responses():
                # Process the extra iobuf reads needed
                while self._extra_reads:
                    self._request_iobuf_region_tail(self._extra_reads.pop())

                # Process the next iobuf reads needed
                while self._next_reads:
                    self._request_iobuf_region(self._next_reads.pop())

        for core_subset in core_subsets:
            x, y = core_subset.x, core_subset.y
            for p in core_subset.processor_ids:
                iobuf = ""
                for item in self._iobuf[x, y, p].values():
                    iobuf += item.decode(_ENCODING)
                yield IOBuffer(x, y, p, iobuf)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from contextlib import suppress
import logging
import functools
from os.path import join
from typing import Dict, List, Optional, Set, Tuple, cast

from spinn_utilities.config_holder import (
    get_config_bool, get_config_int_or_none, get_config_str_or_none)
from spinn_utilities.data import UtilsDataView
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.typing.coords import XY

from spinn_machine import (Router, Chip, Link, Machine)
from spinn_machine.ignores import IgnoreChip, IgnoreCore, IgnoreLink
from spinn_machine.machine_factory import machine_repair

from spinnman.constants import (
    ROUTER_REGISTER_P2P_ADDRESS, SYSTEM_VARIABLE_BASE_ADDRESS)
from spinnman.data import SpiNNManDataView
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.impl import ReadMemory, ReadLink, GetChipInfo
from spinnman.messages.scp.impl.get_chip_info_response import (
    GetChipInfoResponse)
from spinnman.messages.scp.impl.read_memory import Response
from spinnman.messages.spinnaker_boot import (
    SystemVariableDefinition)
from spinnman.model import ChipSummaryInfo, P2PTable
from spinnman.model.enums import CPUState

from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import \
    ConnectionSelector

logger = FormatAdapter(logging.getLogger(__name__))

REPORT_FILE = "Ignores_report.rpt"

P_TO_V = SystemVariableDefinition.physical_to_virtual_core_map
V_TO_P = SystemVariableDefinition.virtual_to_physical_core_map
P_TO_V_ADDR = SYSTEM_VARIABLE_BASE_ADDRESS + P_TO_V.offset
_P_TO_V_SIZE = cast(int, P_TO_V.array_size)
P_MAPS_SIZE = _P_TO_V_SIZE + cast(int, V_TO_P.array_size)


class GetMachineProcess(AbstractMultiConnectionProcess):
    """
    A process for getting the machine details over a set of connections.
    """
    __slots__ = (
        "_chip_info",
        # Used if there are any ignores with IP addresses
        # Holds a mapping from IP to board root (x,y)
        "_ethernets",
        # Holds a map from x,y to a set of virtual cores to ignores
        "_ignore_cores_map",
        "_p2p_column_data",
        # Holds a mapping from (x,y) to a mapping of physical to virtual core
        "_virtual_to_physical_map",
        # Holds a mapping from (x,y) to a mapping of virtual to physical core
        "_physical_to_virtual_map",
        # Progress bar to fill in as details are received
        "_progress")

    def __init__(self, connection_selector: ConnectionSelector):
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector)

        self._ignore_cores_map: Dict[XY, Set[int]] = defaultdict(set)

        self._p2p_column_data: List[Tuple[bytes, int]] = list()

        # A dictionary of (x, y) -> ChipInfo
        self._chip_info: Dict[XY, ChipSummaryInfo] = dict()

        # Set to None meaning not computed yet
        self._ethernets: Optional[Dict[str, XY]] = None

        # Maps between virtual and physical cores
        self._virtual_to_physical_map: Dict[XY, bytes] = dict()
        self._physical_to_virtual_map: Dict[XY, bytes] = dict()
        self._progress: Optional[ProgressBar] = None

    def _make_chip(self, chip_info: ChipSummaryInfo, machine: Machine) -> Chip:
        """
        Creates a chip from a ChipSummaryInfo structure.

        :param ChipSummaryInfo chip_info:
            The ChipSummaryInfo structure to create the chip from
        :return: The created chip
        :rtype: ~spinn_machine.Chip
        """
        # Keep track of progress
        self._progress = None
        # Create the down cores set if any
        n_cores = \
            SpiNNManDataView.get_machine_version().max_cores_per_chip
        n_cores = min(chip_info.n_cores, n_cores)
        core_states = chip_info.core_states
        down_cores = self._ignore_cores_map.get(
            (chip_info.x, chip_info.y), set())
        if 0 in down_cores:
            raise NotImplementedError(
                "Declaring scamp core (0) as down is not supported")
        cores = list()
        for i in range(1, n_cores):
            if i in down_cores:
                pass
            elif core_states[i] != CPUState.IDLE:
                self._report_ignore(
                    "Not using core {}, {}, {} in state {}",
                    chip_info.x, chip_info.y, i,  core_states[i])
            else:
                cores.append(i)

        # Create the router
        router = self._make_router(chip_info, machine)

        # Create the chip's SDRAM object
        sdram_size = chip_info.largest_free_sdram_block
        max_sdram_size = get_config_int_or_none(
            "Machine", "max_sdram_allowed_per_chip")
        if (max_sdram_size is not None and
                sdram_size > max_sdram_size):
            sdram_size = max_sdram_size

        # Create the chip
        return Chip(
            x=chip_info.x, y=chip_info.y, scamp_processors=[0],
            placable_processors=cores,
            router=router, sdram=sdram_size,
            ip_address=chip_info.ethernet_ip_address,
            nearest_ethernet_x=chip_info.nearest_ethernet_x,
            nearest_ethernet_y=chip_info.nearest_ethernet_y,
            parent_link=chip_info.parent_link)

    def _make_router(
            self, chip_info: ChipSummaryInfo, machine: Machine) -> Router:
        """
        :param ChipSummaryInfo chip_info:
        :param ~spinn_machine.Machine machine:
        :rtype: ~spinn_machine.Router
        """
        links = list()
        for link in chip_info.working_links:
            dest_xy = machine.xy_over_link(chip_info.x, chip_info.y, link)
            if dest_xy in self._chip_info:
                links.append(Link(
                    chip_info.x, chip_info.y, link, dest_xy[0], dest_xy[1]))
            else:
                logger.warning(
                    "Link {},{}:{} points to {} but that is not included ",
                    chip_info.x, chip_info.y, link, dest_xy)

        return Router(
            links=links,
            n_available_multicast_entries=(
                chip_info.n_free_multicast_routing_entries))

    def __receive_p2p_data(
            self, column: int, scp_read_response: Response):
        """
        :param int column:
        :param Response scp_read_response:
        """
        self._p2p_column_data[column] = (
            scp_read_response.data, scp_read_response.offset)

    def _receive_chip_info(
            self, scp_read_chip_info_response: GetChipInfoResponse):
        """
        :param GetChipInfoResponse scp_read_chip_info_response:
        """
        chip_info = scp_read_chip_info_response.chip_info
        self._chip_info[chip_info.x, chip_info.y] = chip_info
        if self._progress is not None:
            self._progress.update()

    def _receive_p_maps(
            self, x: int, y: int, scp_read_response: Response):
        """
        Receive the physical-to-virtual and virtual-to-physical maps.

        :param Response scp_read_response:
        """
        data = scp_read_response.data
        off = scp_read_response.offset
        self._physical_to_virtual_map[x, y] = data[off:_P_TO_V_SIZE + off]
        off += _P_TO_V_SIZE
        self._virtual_to_physical_map[x, y] = data[off:]

    def _receive_error(
            self, request: AbstractSCPRequest, exception, tb, connection):
        """
        :param AbstractSCPRequest request:
        :param Exception exception:
        """
        # If we get an ReadLink with a
        # SpinnmanUnexpectedResponseCodeException, this is a failed link
        # and so can be ignored
        if isinstance(request, ReadLink):
            if isinstance(exception, SpinnmanUnexpectedResponseCodeException):
                return
        super()._receive_error(request, exception, tb, connection)

    def get_machine_details(
            self, boot_x: int, boot_y: int,
            width: int, height: int) -> Machine:
        """
        :param int boot_x:
        :param int boot_y:
        :param int width:
        :param int height:
        :rtype: ~spinn_machine.Machine
        """
        # Get the P2P table - 8 entries are packed into each 32-bit word
        p2p_column_bytes = P2PTable.get_n_column_bytes(height)
        blank = (b'', 0)
        self._p2p_column_data = [blank] * width
        with self._collect_responses():
            for column in range(width):
                offset = P2PTable.get_column_offset(column)
                self._send_request(
                    ReadMemory(
                        coordinates=(boot_x, boot_y, 0),
                        base_address=(ROUTER_REGISTER_P2P_ADDRESS + offset),
                        size=p2p_column_bytes),
                    functools.partial(self.__receive_p2p_data, column))
        p2p_table = P2PTable(width, height, self._p2p_column_data)

        # Get the chip information for each chip
        self._progress = ProgressBar(
            p2p_table.n_routes,
            f"Reading details from {p2p_table.n_routes} chips")
        with suppress(Exception), self._collect_responses():
            # Ignore errors, as any error here just means that a chip
            # is down that wasn't marked as down
            for (x, y) in p2p_table.iterchips():
                self._send_request(GetChipInfo(x, y), self._receive_chip_info)
                self._send_request(
                    ReadMemory((x, y, 0), P_TO_V_ADDR, P_MAPS_SIZE),
                    functools.partial(self._receive_p_maps, x, y))
        self._progress.end()
        SpiNNManDataView.set_v_to_p_map(self._virtual_to_physical_map)

        # Warn about unexpected missing chips
        for (x, y) in p2p_table.iterchips():
            if (x, y) not in self._chip_info:
                logger.warning(
                    "Chip {}, {} was expected but didn't reply", x, y)

        version = SpiNNManDataView.get_machine_version()
        machine = version.create_machine(width, height)
        self._preprocess_ignore_chips(machine)
        self._process_ignore_links(machine)
        self._preprocess_ignore_cores(machine)

        return self._fill_machine(machine)

    def _fill_machine(self, machine: Machine) -> Machine:
        """
        :param ~spinn_machine.Machine machine:
        :rtype: ~spinn_machine.Machine
        """
        for chip_info in sorted(
                self._chip_info.values(), key=lambda chip: (chip.x, chip.y)):
            if (chip_info.ethernet_ip_address is not None and
                    (chip_info.x != chip_info.nearest_ethernet_x
                     or chip_info.y != chip_info.nearest_ethernet_y)):
                if get_config_bool("Machine", "ignore_bad_ethernets"):
                    logger.warning(
                        "Chip {}:{} claimed it has ip address: {}. "
                        "This ip will not be used.",
                        chip_info.x, chip_info.y,
                        chip_info.ethernet_ip_address)
                    chip_info.clear_ethernet_ip_address()
                else:
                    logger.warning(
                        "Not using chip {}:{} as it has an unexpected "
                        "ip address: {}", chip_info.x, chip_info.y,
                        chip_info.ethernet_ip_address)
                    continue

            # If the above has not continued, add the chip
            machine.add_chip(self._make_chip(chip_info, machine))

        machine.validate()
        return machine_repair(machine)

    # Stuff below here is purely for dealing with ignores

    def _process_ignore_links(self, machine: Machine):
        """
        Processes the collection of ignore links to remove then from chip info.

        Converts any local (x, y, IP address) to global (x, y)

        Discards any ignores which are known to have no
        affect based on the already read chip_info

        Also removes any inverse links

        Logs all actions except for ignores with unused IP addresses

        :param ~spinn_machine.Machine machine:
            An empty machine to handle wrap-arounds
        """
        for ignore in IgnoreLink.parse_string(
                get_config_str_or_none("Machine", "down_links")):
            global_xy = self._ignores_local_to_global(
                ignore.x, ignore.y, ignore.ip_address, machine)
            if global_xy is None:
                continue
            chip_info = self._chip_info.get(global_xy, None)
            if chip_info is None:
                self._report_ignore(
                    "Discarding ignore link on chip {} as it is not/ no longer"
                    " in info", global_xy)
                continue
            link = ignore.link
            if link in chip_info.working_links:
                chip_info.working_links.remove(link)
                self._report_ignore(
                    "On chip {} ignoring link:{}", global_xy, link)
                # ignore the inverse link too
                inv_xy = machine.xy_over_link(global_xy[0], global_xy[1], link)
                if inv_xy in self._chip_info:
                    inv_chip_info = self._chip_info[inv_xy]
                    inv_link = (link + 3) % 6
                    if inv_link in inv_chip_info.working_links:
                        inv_chip_info.working_links.remove(inv_link)
                        self._report_ignore(
                            "On chip {} ignoring link {} as it is the inverse "
                            "of link {} on chip {}", inv_xy, inv_link, link,
                            global_xy)
            else:
                self._report_ignore(
                    "Discarding ignore link {} on chip {} as it is not/"
                    "no longer in info", link, global_xy)

    def _preprocess_ignore_cores(self, machine: Machine):
        """
        Converts the collection of ignore cores into a map of ignore by x,y.

        Converts any local (x, y, IP address) to global (x, y)

        Discards (with log messages) any ignores which are known to have no
        affect based on the already read chip_info

        Converts any physical  cores to virtual ones.
        Core numbers <= 0 are assumed to be 0 - physical_id

        :param ~spinn_machine.Machine machine:
            An empty machine to handle wrap-arounds
        """
        # Convert by IP to global
        for ignore in IgnoreCore.parse_string(
                get_config_str_or_none("Machine", "down_cores")):
            global_xy = self._ignores_local_to_global(
                ignore.x, ignore.y, ignore.ip_address, machine)
            if global_xy is None:
                continue
            p = self._get_virtual_p(global_xy, ignore.p)
            if p is not None:
                self._ignore_cores_map[global_xy].add(p)

    def _preprocess_ignore_chips(self, machine: Machine):
        """
        Processes the collection of ignore chips and discards their chip info.

        Converts any local (x, y IP address) to global (x, y)

        Discards any ignores which are known to have no
        affect based on the already read chip_info

        Logs all actions except for ignores with unused IP addresses

        :param ~spinn_machine.Machine machine:
            An empty machine to handle wrap-arounds
        """
        for ignore in IgnoreChip.parse_string(
                get_config_str_or_none("Machine", "down_chips")):
            # Convert by IP to global
            global_xy = self._ignores_local_to_global(
                ignore.x, ignore.y, ignore.ip_address, machine)
            if global_xy is None:
                continue  # Never on this machine
            chip_info = self._chip_info.pop(global_xy, None)
            if chip_info is None:
                continue  # Already ignored maybe by a dead chip list
            self._report_ignore("Chip {} will be ignored", global_xy)
            for link in chip_info.working_links:
                # ignore the inverse link
                inv_xy = machine.xy_over_link(global_xy[0], global_xy[1], link)
                if inv_xy in self._chip_info:
                    inv_chip_info = self._chip_info[inv_xy]
                    inv_link = (link + 3) % 6
                    if inv_link in inv_chip_info.working_links:
                        inv_chip_info.working_links.remove(inv_link)
                        self._report_ignore(
                            "On chip {} ignoring link {} as it points to "
                            "ignored chip chip {}",
                            inv_xy, inv_link, global_xy)

    def _ignores_local_to_global(
            self, local_x: int, local_y: int, ip_address: Optional[str],
            machine: Machine) -> Optional[XY]:
        """
        :param int local_x:
        :param int local_y:
        :param str ip_address:
        :param ~spinn_machine.Machine machine:
        :rtype: tuple(int,int)
        """
        if ip_address is None:
            global_xy = (local_x, local_y)
        else:
            ethernet = self._ethernet_by_ipaddress(ip_address)
            if ethernet is None:
                self._report_ignore(
                    "Ignore with ip:{} will be discarded as no board with "
                    "that ip in this machine", ip_address)
                return None
            global_xy = machine.get_global_xy(
                local_x, local_y, ethernet[0], ethernet[1])
            self._report_ignore(
                "Ignores for local x:{} y:{} and ip:{} map to global {} with "
                "root {}", local_x, local_y, ip_address, global_xy,
                self._chip_info[(0, 0)].ethernet_ip_address)

        if global_xy in self._chip_info:
            return global_xy
        else:
            self._report_ignore(
                "Ignore for global chip {} will be discarded as no such chip "
                "in this machine", global_xy)
            return None

    def _ethernet_by_ipaddress(self, ip_address: str) -> Optional[XY]:
        """
        :param str ip_address:
        :rtype: tuple(int,int)
        """
        if self._ethernets is None:
            self._ethernets = dict()
            for chip_info in self._chip_info.values():
                if chip_info.ethernet_ip_address is not None:
                    self._ethernets[chip_info.ethernet_ip_address] = \
                        (chip_info.x, chip_info.y)
        return self._ethernets.get(ip_address, None)

    def _get_virtual_p(self, xy: XY, p: int) -> Optional[int]:
        """
        :param tuple(int,int) xy:
        :param int p:
        :rtype: int
        """
        if p > 0:
            self._report_ignore("On chip {} ignoring core {}", xy, p)
            return p
        virtual_map = self._physical_to_virtual_map[xy]
        physical = 0 - p
        if physical >= len(virtual_map) or virtual_map[physical] == 0xFF:
            self._report_ignore(
                "On chip {} physical core {} was not used "
                "so ignore is being discarded.", xy, physical)
            return None
        virtual_p = virtual_map[physical]
        if virtual_p == 0:
            self._report_ignore(
                "On chip {} physical core {} was used as the monitor "
                "so will NOT be ignored", xy, physical)
            return None
        self._report_ignore(
            "On chip {} ignoring core {} as it maps to physical "
            "core {}", xy, virtual_p, physical)
        return virtual_p

    def _report_ignore(self, message: str, *args):
        """
        Writes the ignore message by either creating or appending the report.

        The implementation choice to reopen the file every time is not the
        fastest but is the cleanest and safest for code that in default
        conditions is never run.

        :param str message:
        """
        full_message = message.format(*args) + "\n"
        report_file = join(UtilsDataView.get_run_dir_path(), REPORT_FILE)
        with open(report_file, "a", encoding="utf-8") as r_file:
            r_file.write(full_message)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import socket
import select
from contextlib import suppress
from typing import Callable, Optional, Tuple
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_utilities.ping import Ping
from spinnman.exceptions import (SpinnmanIOException, SpinnmanEOFException)
from spinnman.connections.abstract_classes import Connection
from spinnman.utilities.socket_utils import (
    bind_socket, connect_socket, get_udp_socket, get_socket_address,
    resolve_host, set_receive_buffer_size, receive_message,
    receive_message_and_address, send_message, send_message_to_address)
from spinnman.connections.abstract_classes import Listenable

logger = FormatAdapter(logging.getLogger(__name__))
_RECEIVE_BUFFER_SIZE = 1048576
_PING_COUNT = 5
_REPR_TEMPLATE = "UDPConnection(local={}:{}, remote={}:{})"
_MSG_MAX = 300


class UDPConnection(Connection, Listenable[bytes]):
    """
    A connection that routes messages via UDP to some remote host.

    Subclasses of this should be aware that UDP messages may be dropped,
    reordered, or duplicated, and that there's no way to tell whether the
    other end of the connection truly exists except by listening for
    occasional messages from them. There is also an upper size limit on
    messages, formally of 64kB, but usually of about 1500 bytes (the typical
    maximum size of an Ethernet packet);
    SDP messages have lower maximum lengths.
    """

    __slots__ = (
        "_can_send",
        "_local_ip_address",
        "_local_port",
        "_remote_ip_address",
        "_remote_port",
        "_socket")

    def __init__(
            self, local_host: Optional[str] = None,
            local_port: Optional[int] = None,
            remote_host: Optional[str] = None,
            remote_port: Optional[int] = None):
        """
        :param str local_host: The local host name or IP address to bind to.
            If not specified defaults to bind to all interfaces, unless
            remote_host is specified, in which case binding is done to the
            IP address that will be used to send packets
        :param int local_port: The local port to bind to, between 1025 and
            65535. If not specified, defaults to a random unused local port
        :param str remote_host: The remote host name or IP address to send
            packets to. If not specified, the socket will be available for
            listening only, and will throw and exception if used for sending
        :param int remote_port: The remote port to send packets to.  If
            remote_host is None, this is ignored.  If remote_host is specified
            specified, this must also be specified for the connection to allow
            sending
        :raise SpinnmanIOException:
            If there is an error setting up the communication channel
        """
        self._socket = get_udp_socket()
        set_receive_buffer_size(self._socket, _RECEIVE_BUFFER_SIZE)

        # Get the host and port to bind to locally
        local_bind_host = "" if local_host is None else local_host
        local_bind_port = 0 if local_port is None else local_port
        bind_socket(self._socket, local_bind_host, local_bind_port)

        # Mark the socket as non-sending, unless the remote host is
        # specified - send requests will then cause an exception
        self._can_send = False
        self._remote_ip_address: Optional[str] = None
        self._remote_port: Optional[int] = None

        # Get the host to connect to remotely
        if remote_host is not None and remote_port is not None:
            self._remote_port = remote_port
            self._remote_ip_address = resolve_host(remote_host)
            connect_socket(self._socket, self._remote_ip_address, remote_port)
            self._can_send = True

        # If things are closed here, it's a catastrophic problem
        if self.__is_closed:
            raise SpinnmanEOFException()

        # Get the details of where the socket is connected
        self._local_ip_address, self._local_port = \
            get_socket_address(self._socket)

        # Set a general timeout on the socket
        self._socket.settimeout(1.0)

    @property
    def __is_closed(self) -> bool:
        """
        Is the socket closed?

        .. note::
            Just because a socket is not closed doesn't mean that you're going
            to be able to successfully write to it or read from it; some
            failures are only detected on use. But closed sockets definitely
            behave in certain ways!

        :rtype: bool
        """
        # Reach into Python'#s guts
        # pylint: disable=protected-access
        return self._socket._closed  # type: ignore[attr-defined]

    @overrides(Connection.is_connected)
    def is_connected(self) -> bool:
        # Closed sockets are never connected!
        if self.__is_closed:
            return False

        # If this is not a sending socket, it is not connected
        if not self._can_send:
            return False

        # check if machine is active and on the network
        for _ in range(_PING_COUNT):
            # Assume connected if ping works
            if Ping.ping(self._remote_ip_address) == 0:
                return True

        # If the ping fails this number of times, the host cannot be contacted
        return False

    @property
    def local_ip_address(self) -> str:
        """
        The local IP address to which the connection is bound,
        as a dotted string, e.g., `0.0.0.0`.

        :rtype: str
        """
        return self._local_ip_address

    @property
    def local_port(self) -> int:
        """
        The number of the local port to which the connection is bound.

        :rtype: int
        """
        return self._local_port

    @property
    def remote_ip_address(self) -> Optional[str]:
        """
        The remote IP address to which the connection is connected,
        or `None` if not connected remotely.

        :rtype: str
        """
        return self._remote_ip_address

    @property
    def remote_port(self) -> Optional[int]:
        """
        The remote port number to which the connection is connected,
        or `None` if not connected remotely.

        :rtype: int
        """
        return self._remote_port

    def receive(self, timeout: Optional[float] = None) -> bytes:
        """
        Receive data from the connection.

        :param float timeout: The timeout in seconds, or `None` to wait forever
        :return: The data received as a byte-string
        :rtype: bytes
        :raise SpinnmanTimeoutException:
            If a timeout occurs before any data is received
        :raise SpinnmanIOException: If an error occurs receiving the data
        """
        if self.__is_closed:
            raise SpinnmanEOFException()
        return receive_message(self._socket, timeout, _MSG_MAX)

    def receive_with_address(self, timeout: Optional[float] = None) -> Tuple[
            bytes, Tuple[str, int]]:
        """
        Receive data from the connection along with the address where the
        data was received from.

        :param float timeout: The timeout, or `None` to wait forever
        :return: A tuple of the data received and a tuple of the
            (address, port) received from
        :rtype: tuple(bytes, tuple(str, int))
        :raise SpinnmanTimeoutException:
            If a timeout occurs before any data is received
        :raise SpinnmanIOException: If an error occurs receiving the data
        """
        if self.__is_closed:
            raise SpinnmanEOFException()
        return receive_message_and_address(self._socket, timeout, _MSG_MAX)

    def send(self, data: bytes):
        """
        Send data down this connection.

        :param data: The data to be sent
        :type data: bytes or bytearray
        :raise SpinnmanIOException: If there is an error sending the data
        """
        if self.__is_closed:
            raise SpinnmanEOFException()
        if not self._can_send:
            raise SpinnmanIOException(
                "Remote host and/or port not set - data cannot be sent with"
                " this connection")
        while not send_message(self._socket, data):
            if self.__is_closed:
                raise SpinnmanEOFException()

    def send_to(self, data: bytes, address: Tuple[str, int]):
        """
        Send data down this connection.

        :param data: The data to be sent as a byte-string
        :type data: bytes or bytearray
        :param tuple(str,int) address:
            A tuple of (address, port) to send the data to
        :raise SpinnmanIOException: If there is an error sending the data
        """
        if self.__is_closed:
            raise SpinnmanEOFException()
        while not send_message_to_address(self._socket, data, address):
            if self.__is_closed:
                raise SpinnmanEOFException()

    @overrides(Connection.close)
    def close(self) -> None:
        if self.__is_closed:
            return
        with suppress(Exception):
            self._socket.shutdown(socket.SHUT_WR)
        self._socket.close()

    def is_ready_to_receive(self, timeout: float = 0):
        if self.__is_closed:
            return True
        return len(select.select([self._socket], [], [], timeout)[0]) == 1

    def __repr__(self) -> str:
        return _REPR_TEMPLATE.format(
            self.local_ip_address, self.local_port,
            self.remote_ip_address, self.remote_port)

    @overrides(Listenable.get_receive_method)
    def get_receive_method(self) -> Callable[[], bytes]:
        return self.receive



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


# pylint: disable=invalid-name
class SDP_RUNNING_MESSAGE_CODES(Enum):
    """
    Codes for sending control messages to spin1_api.
    """
    SDP_STOP_ID_CODE = 6
    SDP_NEW_RUNTIME_ID_CODE = 7
    SDP_UPDATE_PROVENCE_REGION_AND_EXIT = 8
    SDP_CLEAR_IOBUF_CODE = 9



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from spinn_utilities.overrides import overrides
from spinnman.messages.scp.abstract_messages import AbstractSCPResponse
from spinnman.messages.scp.enums import SCPResult
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException

_BYTE_SKIP_BYTE_BYTE = struct.Struct("<Bx2B")


class IPTagGetInfoResponse(AbstractSCPResponse):
    """
    An SCP response to a request for information about IP tags.
    """
    __slots__ = (
        "_fixed_size",
        "_pool_size",
        "_tto")

    def __init__(self) -> None:
        super().__init__()
        self._tto = 0
        self._pool_size = 0
        self._fixed_size = 0

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "Get IP Tag Info", "CMD_IPTAG", result.name)

        self._tto, self._pool_size, self._fixed_size = \
            _BYTE_SKIP_BYTE_BYTE.unpack_from(data, offset)

    @property
    def transient_timeout(self) -> int:
        """
        The timeout for transient IP tags (i.e. responses to SCP commands).

        :rtype: int
        """
        return self._tto

    @property
    def pool_size(self) -> int:
        """
        The count of the IP tag pool size.

        :rtype: int
        """
        return self._pool_size

    @property
    def fixed_size(self) -> int:
        """
        The count of the number of fixed IP tag entries.

        :rtype: int
        """
        return self._fixed_size



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Union, TYPE_CHECKING
from spinn_utilities.abstract_base import abstractmethod, AbstractBase
if TYPE_CHECKING:
    from spinnman.messages.eieio.command_messages import EIEIOCommandHeader
    from spinnman.messages.eieio.data_messages import EIEIODataHeader


class AbstractEIEIOMessage(object, metaclass=AbstractBase):
    """
    Interface for an EIEIOMessage.
    """
    __slots__ = ()

    @property
    @abstractmethod
    def eieio_header(self) -> Union[EIEIOCommandHeader, EIEIODataHeader]:
        """
        The header of the message.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def bytestring(self) -> bytes:
        """
        The bytes of the message.

        :rtype: bytes
        """
        raise NotImplementedError



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinnman.messages.scp.impl.sdram_alloc import SDRAMAlloc, _AllocResponse
from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)


class MallocSDRAMProcess(AbstractMultiConnectionProcess[_AllocResponse]):
    """
    A process for allocating a block of SDRAM on a SpiNNaker chip.
    """
    __slots__ = ("_base_address", )

    def __init__(self, connection_selector: ConnectionSelector):
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector)
        self._base_address = 0

    def __handle_sdram_alloc_response(self, response: _AllocResponse):
        self._base_address = response.base_address

    def malloc_sdram(self, x: int, y: int, size: int, app_id: int, tag: int):
        """
        Allocate space in the SDRAM space.

        :param int x:
        :param int y:
        :param int size:
        :param int app_id:
        :param int tag:
        """
        # pylint: disable=too-many-arguments
        with self._collect_responses():
            self._send_request(SDRAMAlloc(x, y, app_id, size, tag),
                               self.__handle_sdram_alloc_response)

    @property
    def base_address(self) -> int:
        """
        The address of the allocated memory block.

        :rtype: int
        """
        return self._base_address



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Optional, Tuple
from spinn_utilities.abstract_base import AbstractBase
from spinn_utilities.overrides import overrides
from spinnman.connections.udp_packet_connections import SCAMPConnection
from spinnman.messages.sdp import SDPMessage, SDPFlag
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPResult
from .spalloc_proxied_connection import SpallocProxiedConnection

_TWO_SHORTS = struct.Struct("<2H")
_TWO_SKIP: bytes = b'\0\0'


class SpallocSCPConnection(
        SCAMPConnection, SpallocProxiedConnection,
        metaclass=AbstractBase):
    """
    The socket interface supported by proxied sockets. The socket will always
    be talking to a specific board. This emulates a
    :py:class:`SCAMPConnection`.
    """
    __slots__ = ()

    def __init__(self, x, y):
        super(SpallocSCPConnection, self).__init__(x, y)

    @overrides(SCAMPConnection.receive_sdp_message)
    def receive_sdp_message(
            self, timeout: Optional[float] = None) -> SDPMessage:
        data = self.receive(timeout)
        return SDPMessage.from_bytestring(data, 2)

    @overrides(SCAMPConnection.send_sdp_message)
    def send_sdp_message(self, sdp_message: SDPMessage):
        # If a reply is expected, the connection should
        if sdp_message.sdp_header.flags == SDPFlag.REPLY_EXPECTED:
            sdp_message.sdp_header.update_for_send(self.chip_x, self.chip_y)
        else:
            sdp_message.sdp_header.update_for_send(0, 0)
        self.send(_TWO_SKIP + sdp_message.bytestring)

    @overrides(SCAMPConnection.receive_scp_response)
    def receive_scp_response(self, timeout: Optional[float] = 1.0) -> Tuple[
            SCPResult, int, bytes, int]:
        data = self.receive(timeout)
        result, sequence = _TWO_SHORTS.unpack_from(data, 10)
        return SCPResult(result), sequence, data, 2

    @overrides(SCAMPConnection.get_scp_data)
    def get_scp_data(
            self, scp_request: AbstractSCPRequest,
            x: Optional[int] = None, y: Optional[int] = None) -> bytes:
        if x is None:
            x = self.chip_x
        if y is None:
            y = self.chip_y
        scp_request.sdp_header.update_for_send(x, y)
        return _TWO_SKIP + scp_request.bytestring



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from spinn_utilities.overrides import overrides
from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.messages.eieio import EIEIOType
from .abstract_data_element import AbstractDataElement

_TWO_SHORTS = struct.Struct("<HH")
_TWO_WORDS = struct.Struct("<II")


class KeyPayloadDataElement(AbstractDataElement):
    """
    A data element that contains a key and a payload.
    """
    __slots__ = (
        "_key",
        "_payload",
        "_payload_is_timestamp")

    def __init__(
            self, key: int, payload: int, payload_is_timestamp: bool = False):
        self._key = key
        self._payload = payload
        self._payload_is_timestamp = payload_is_timestamp

    @property
    def key(self) -> int:
        """
        The key value passed into the init.

        :rtype: int
        """
        return self._key

    @property
    def payload(self) -> int:
        """
        Gets the payload value passed into the init.

        :rtype: int
        """
        return self._payload

    @property
    def payload_is_timestamp(self) -> bool:
        """
        Gets the payload_is_timestamp passed into the init.

        :rtype: int
        """
        return self._payload_is_timestamp

    @overrides(AbstractDataElement.get_bytestring)
    def get_bytestring(self, eieio_type: EIEIOType) -> bytes:
        if eieio_type.payload_bytes == 0:
            raise SpinnmanInvalidParameterException(
                "eieio_type", eieio_type,
                "The type specifies no payload, but this element has a "
                "payload")
        if eieio_type == EIEIOType.KEY_PAYLOAD_16_BIT:
            return _TWO_SHORTS.pack(self._key, self._payload)
        elif eieio_type == EIEIOType.KEY_PAYLOAD_32_BIT:
            return _TWO_WORDS.pack(self._key, self._payload)
        else:
            raise SpinnmanInvalidParameterException(
                "eieio_type", eieio_type, "Unknown type")

    def __str__(self):
        return f"KeyPayloadDataElement:0x{self._key:x}:0x{self._payload:x}"

    def __repr__(self):
        return self.__str__()



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import struct
from time import localtime, asctime
from typing import Tuple, cast
from typing_extensions import TypeAlias
from spinnman.exceptions import SpinnmanInvalidParameterException

_VERSION_PATTERN = struct.Struct("<BBBBHHI")
_V: TypeAlias = Tuple[int, int, int]


class VersionInfo(object):
    """
    Decodes SC&MP/SARK version information as returned by the SVER command.
    """
    __slots__ = [
        "_build_date",
        "_hardware",
        "_name",
        "_physical_cpu_id",
        "_version_number",
        "_version_string",
        "_x", "_y", "_p"]

    def __init__(self, version_data: bytes, offset: int = 0):
        """
        :param bytes version_data:
            bytes from an SCP packet containing version information
        :param int offset:
            the offset in the bytes from an SCP packet containing
            version information
        :raise SpinnmanInvalidParameterException:
            If the message does not contain valid version information
        """
        (self._p, self._physical_cpu_id, self._y, self._x, _,
            version_no, self._build_date) = _VERSION_PATTERN.unpack_from(
                memoryview(version_data), offset)

        version_str = version_data[offset + 12:-1].decode("utf-8")

        self._version_number: _V
        if version_no < 0xFFFF:
            try:
                self._version_number = (version_no // 100, version_no % 100, 0)
                self._name, self._hardware = version_str.split("/")
                self._version_string = version_str
            except ValueError as exception:
                raise SpinnmanInvalidParameterException(
                    "version_data", version_str,
                    f"Incorrect format: {exception}") from exception
        else:
            name_hardware, _, version = version_str.partition("\0")
            self._version_string = version
            matches = re.match(r"(\d+)\.(\d+)\.(\d+)", version)
            if matches is None:
                raise SpinnmanInvalidParameterException(
                    "version", version, "Cannot be parsed")
            self._version_number = cast(_V, tuple(
                map(int, matches.group(1, 2, 3))))
            self._name, self._hardware = name_hardware.rstrip("\0").split("/")

    @property
    def name(self) -> str:
        """
        The name of the software.

        :rtype: str
        """
        return self._name

    @property
    def version_number(self) -> _V:
        """
        The version number of the software.

        :rtype: tuple(int, int, int)
        """
        return self._version_number

    @property
    def hardware(self) -> str:
        """
        The hardware being run on.

        :rtype: str
        """
        return self._hardware

    @property
    def x(self) -> int:
        """
        The X-coordinate of the chip where the information was obtained.

        :rtype: int
        """
        return self._x

    @property
    def y(self) -> int:
        """
        The Y-coordinate of the chip where the information was obtained.

        :rtype: int
        """
        return self._y

    @property
    def p(self) -> int:
        """
        The processor ID of the processor where the information was obtained.

        :rtype: int
        """
        return self._p

    @property
    def build_date(self) -> int:
        """
        The build date of the software, in seconds since 1st January 1970.

        :rtype: int
        """
        return self._build_date

    @property
    def version_string(self) -> str:
        """
        The version information as text.

        :rtype: str
        """
        return self._version_string

    def __str__(self) -> str:
        return (f"[Version: {self._name} {self._version_string} at "
                f"{self._hardware}:{self._x}:{self._y}:{self._p} "
                f"(built {asctime(localtime(self._build_date))})]")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-arguments
import logging
from typing import Optional
from threading import Condition, RLock
from spinn_utilities.abstract_base import (
    AbstractBase, abstractmethod)
from spinn_utilities.log import FormatAdapter
from spinnman.connections.udp_packet_connections import BMPConnection
from spinnman.processes import ConnectionSelector, FixedConnectionSelector
from spinnman.transceiver.transceiver import Transceiver

logger = FormatAdapter(logging.getLogger(__name__))


class ExtendableTransceiver(Transceiver, metaclass=AbstractBase):
    """
    Support Functions to allow a Transceiver to also be an ExtendedTransceiver

    If supporting these is too difficult the
    methods that require these may be removed.
    """
    __slots__ = [
        "_flood_write_lock",
        "_nearest_neighbour_id",
        "_nearest_neighbour_lock"
    ]

    def __init__(self):

        # A lock against multiple flood fill writes - needed as SCAMP cannot
        # cope with this
        self._flood_write_lock = Condition()

        # The nearest neighbour start ID and lock
        self._nearest_neighbour_id = 1
        self._nearest_neighbour_lock = RLock()

    @property
    @abstractmethod
    def bmp_selector(self) -> Optional[FixedConnectionSelector[BMPConnection]]:
        """
        Returns the BMP selector

        :rtype: AbstractMultiConnectionProcessConnectionSelector
        """
        raise NotImplementedError("This method is abstract")

    @property
    @abstractmethod
    def scamp_connection_selector(self) -> ConnectionSelector:
        """
        Returns the scamp selector

        :rtype: AbstractMultiConnectionProcessConnectionSelector
        """
        raise NotImplementedError("This method is abstract")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MulticastMessage(object):
    """
    A SpiNNaker Multicast message, comprising a key (determining the target
    locations) and an optional payload.
    """
    __slots__ = (
        "_key",
        "_payload")

    def __init__(self, key, payload=None):
        """
        :param int key: The key of the packet
        :param int payload: The optional payload of the packet
        """
        self._key = key
        self._payload = payload

    @property
    def key(self):
        """
        The key of the packet.

        :rtype: int
        """
        return self._key

    @property
    def payload(self):
        """
        The payload of the packet if there is one, or `None` if there is no
        payload.

        :rtype: int or None
        """
        return self._payload



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from spinn_machine import RoutingEntry
from spinnman.messages.scp.impl.fixed_route_read import (
    FixedRouteRead, _FixedRouteResponse)

from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)


class ReadFixedRouteRoutingEntryProcess(
        AbstractMultiConnectionProcess[_FixedRouteResponse]):
    """
    A process for reading a fixed route routing table entry.
    """

    __slots__ = (
        # the fixed route routing entry from the response
        "_route", )

    def __init__(self, connection_selector: ConnectionSelector):
        """
        :param ConnectionSelector connection_selector:
            the SC&MP connection selector
        """
        super().__init__(connection_selector)
        self._route: Optional[RoutingEntry] = None

    def __handle_read_response(self, response: _FixedRouteResponse):
        self._route = response.route

    def read_fixed_route(
            self, x: int, y: int, app_id: int = 0) -> RoutingEntry:
        """
        Read the fixed route entry installed on a particular chip's router.

        :param int x: The x-coordinate of the chip, between 0 and 255;
            this is not checked due to speed restrictions
        :param int y: The y-coordinate of the chip, between 0 and 255;
            this is not checked due to speed restrictions
        :param int app_id:
            The ID of the application with which to associate the
            routes.  If not specified, defaults to 0.
        :rtype: ~spinn_machine.RoutingEntry
        """
        with self._collect_responses():
            self._send_request(FixedRouteRead(x, y, app_id),
                               self.__handle_read_response)
        assert self._route is not None
        return self._route



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-arguments

from typing import (
    BinaryIO, Collection, Dict, FrozenSet, Iterable,
    List, Optional, Set, Tuple, Union)
from spinn_utilities.overrides import overrides
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.typing.coords import XY
from spinn_machine import (
    CoreSubsets, Machine, MulticastRoutingEntry, RoutingEntry)
from spinn_machine.tags import AbstractTag, IPTag, ReverseIPTag
from spinnman.data import SpiNNManDataView
from spinnman.connections.abstract_classes import Connection
from spinnman.connections.udp_packet_connections import BMPConnection
from spinnman.connections.udp_packet_connections import (
    SCAMPConnection, SDPConnection)
from spinnman.processes import ConnectionSelector, FixedConnectionSelector
from spinnman.messages.scp.enums import Signal
from spinnman.messages.sdp import SDPMessage
from spinnman.model import (
    CPUInfos, DiagnosticFilter, IOBuffer, RouterDiagnostics,
    VersionInfo)
from spinnman.model.enums import CPUState, UserRegister
from spinnman.transceiver.transceiver import Transceiver
from spinnman.transceiver.extendable_transceiver import ExtendableTransceiver
from spinnman.processes import MostDirectConnectionSelector


class MockableTransceiver(ExtendableTransceiver):
    """
    A based for Mock Transceivers
    """
    __slots__ = ["written_memory"]

    def __init__(self) -> None:
        super().__init__()
        self.written_memory: List[
            Tuple[int, int, int, Union[BinaryIO, bytes, int, str],
                  Optional[int], int, int]] = []

    @overrides(Transceiver.send_sdp_message)
    def send_sdp_message(self, message: SDPMessage,
                         connection: Optional[SDPConnection] = None):
        pass

    @overrides(Transceiver.discover_scamp_connections)
    def discover_scamp_connections(self) -> None:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.add_scamp_connections)
    def add_scamp_connections(self, connections: Dict[XY, str]):
        pass

    @overrides(Transceiver.get_machine_details)
    def get_machine_details(self) -> Machine:
        return SpiNNManDataView.get_machine()

    @overrides(Transceiver.get_connections)
    def get_connections(self) -> Set[Connection]:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.get_cpu_infos)
    def get_cpu_infos(
            self, core_subsets: Optional[CoreSubsets] = None,
            states: Union[CPUState, Iterable[CPUState], None] = None,
            include: bool = True) -> CPUInfos:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.get_clock_drift)
    def get_clock_drift(self, x: int, y: int) -> float:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.read_user)
    def read_user(self, x: int, y: int, p: int, user: UserRegister):
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.add_cpu_information_from_core)
    def add_cpu_information_from_core(
            self, cpu_infos: CPUInfos, x: int, y: int, p: int,
            states: Iterable[CPUState]):
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.get_region_base_address)
    def get_region_base_address(self, x: int, y: int, p: int):
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.get_iobuf)
    def get_iobuf(self, core_subsets: Optional[CoreSubsets] = None
                  ) -> Iterable[IOBuffer]:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.get_core_state_count)
    def get_core_state_count(
            self, app_id: int, state: CPUState,
            xys: Optional[Iterable[Tuple[int, int]]] = None) -> int:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.execute_flood)
    def execute_flood(
            self, core_subsets: CoreSubsets,
            executable: Union[BinaryIO, bytes, str], app_id: int, *,
            n_bytes: Optional[int] = None, wait: bool = False):
        pass

    @overrides(Transceiver.read_fpga_register)
    def read_fpga_register(
            self, fpga_num: int, register: int, board: int = 0) -> int:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.write_fpga_register)
    def write_fpga_register(
            self, fpga_num: int, register: int, value: int, board: int = 0):
        pass

    @overrides(Transceiver.read_bmp_version)
    def read_bmp_version(self, board: int) -> VersionInfo:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.write_memory)
    def write_memory(
            self, x: int, y: int, base_address: int,
            data: Union[BinaryIO, bytes, int, str], *,
            n_bytes: Optional[int] = None, offset: int = 0, cpu: int = 0,
            get_sum: bool = False) -> Tuple[int, int]:
        print("Doing write to", x, y)
        self.written_memory.append(
            (x, y, base_address, data, n_bytes, offset, cpu))
        # Hope the return is never used as it will be wrong
        return (-1, -1)

    @overrides(Transceiver.write_user)
    def write_user(
            self, x: int, y: int, p: int, user: UserRegister, value: int):
        pass

    @overrides(Transceiver.read_memory)
    def read_memory(
            self, x: int, y: int, base_address: int, length: int,
            cpu: int = 0) -> bytearray:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.read_word)
    def read_word(
            self, x: int, y: int, base_address: int, cpu: int = 0) -> int:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.stop_application)
    def stop_application(self, app_id: int):
        pass

    @overrides(Transceiver.wait_for_cores_to_be_in_state)
    def wait_for_cores_to_be_in_state(
            self, all_core_subsets: CoreSubsets, app_id: int,
            cpu_states: Union[CPUState, Iterable[CPUState]], *,
            timeout: Optional[float] = None,
            time_between_polls: float = 0.1,
            error_states: FrozenSet[CPUState] = frozenset((
                CPUState.RUN_TIME_EXCEPTION, CPUState.WATCHDOG)),
            counts_between_full_check: int = 100,
            progress_bar: Optional[ProgressBar] = None):
        pass

    @overrides(Transceiver.send_signal)
    def send_signal(self, app_id: int, signal: Signal):
        pass

    @overrides(Transceiver.set_ip_tag)
    def set_ip_tag(self, ip_tag: IPTag, use_sender: bool = False):
        pass

    @overrides(Transceiver.set_reverse_ip_tag)
    def set_reverse_ip_tag(self, reverse_ip_tag: ReverseIPTag):
        pass

    @overrides(Transceiver.clear_ip_tag)
    def clear_ip_tag(self, tag: int, board_address: Optional[str] = None):
        pass

    @overrides(Transceiver.get_tags)
    def get_tags(self, connection: Optional[SCAMPConnection] = None
                 ) -> Iterable[AbstractTag]:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.malloc_sdram)
    def malloc_sdram(
            self, x: int, y: int, size: int, app_id: int, tag: int = 0) -> int:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.load_multicast_routes)
    def load_multicast_routes(
            self, x: int, y: int, routes: Collection[MulticastRoutingEntry],
            app_id: int):
        pass

    @overrides(Transceiver.load_fixed_route)
    def load_fixed_route(
            self, x: int, y: int, fixed_route: RoutingEntry, app_id: int):
        pass

    @overrides(Transceiver.read_fixed_route)
    def read_fixed_route(self, x: int, y: int, app_id: int) -> RoutingEntry:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.get_multicast_routes)
    def get_multicast_routes(
            self, x: int, y: int,
            app_id: Optional[int] = None) -> List[MulticastRoutingEntry]:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.clear_multicast_routes)
    def clear_multicast_routes(self, x: int, y: int):
        pass

    @overrides(Transceiver.get_router_diagnostics)
    def get_router_diagnostics(self, x: int, y: int) -> RouterDiagnostics:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.get_scamp_connection_selector)
    def get_scamp_connection_selector(self) -> MostDirectConnectionSelector:
        raise NotImplementedError("Needs to be mocked")

    @overrides(Transceiver.set_router_diagnostic_filter)
    def set_router_diagnostic_filter(
            self, x: int, y: int, position: int,
            diagnostic_filter: DiagnosticFilter):
        pass

    @overrides(Transceiver.clear_router_diagnostic_counters)
    def clear_router_diagnostic_counters(self, x: int, y: int):
        pass

    @overrides(Transceiver.close)
    def close(self) -> None:
        pass

    @overrides(Transceiver.control_sync)
    def control_sync(self, do_sync: bool):
        pass

    @overrides(Transceiver.update_provenance_and_exit)
    def update_provenance_and_exit(self, x: int, y: int, p: int):
        pass

    @overrides(Transceiver.send_chip_update_provenance_and_exit)
    def send_chip_update_provenance_and_exit(self, x: int, y: int, p: int):
        pass

    @property
    @overrides(ExtendableTransceiver.bmp_selector)
    def bmp_selector(self) -> Optional[FixedConnectionSelector[BMPConnection]]:
        raise NotImplementedError("Needs to be mocked")

    @property
    @overrides(ExtendableTransceiver.scamp_connection_selector)
    def scamp_connection_selector(self) -> ConnectionSelector:
        raise NotImplementedError("Needs to be mocked")
